#
# An asynchronous I/O library.
#
# author: asfandyar qureshi (may 2006)
#
from select import select
from time import time

from traceback import format_exc
from sys import stderr

class SelectLoop:
    """
    A wrapper for an infinite loop wrapper around 'select'.

    Aids in asynchronous I/O for file descriptors, and
    timeouts.

    HANDLERS:

    I/O handlers:
       An I/O handler object passed to this class must
       be callable with two arguments (fd, time_now).
         fd: the file descriptor on which the I/O event occured.
         time_now: the current time as returned by time.time()
                  
    Timeout handlers:
       A timeout handler object passed to this class must
       be callable with a single argument: the current time.    
    """
    class _Timeout:
        def __init__(self, time_fire_sec, timeout_sec, timeout_handler, run_once=True):
            self.time_fire_sec = time_fire_sec
            self.timeout_sec = timeout_sec
            self.timeout_handler = timeout_handler
            self.run_once = run_once

        def __str__(self):
            return "fire:" + str(self.time_fire_sec) + ", timeout:" + str(self.timeout_sec)
    
    def __init__(self):
        # read/write/exception sets
        self._read_set = []
        self._write_set = []
        self._x_set = []

        # dict: int -> function
        self._read_handlers = {}
        self._write_handlers = {}

        # list of timeouts and handlers, sorted so that
        # the earliest timeout is element zero
        # list: _Timeout
        self._timeouts = []

        self._running = False

    #
    # Read/Write handlers
    #
    
    def want_read(self, read_fd, read_handler_function):
        """
        read_handler_function gets called when read_fd is
        readable.        
        """
        self._read_set.append(read_fd)
        self._read_handlers[read_fd] = read_handler_function

    def want_write(self, write_fd, write_handler_function):
        """
        write_handler_function gets called when write_fd is
        writable.
        """
        self._read_set.append(read_fd)
        self._read_handlers[read_fd] = read_handler_function

    def ignore_read(self, read_fd):
        """
        subsequent read-available events on read_fd ignored.
        """
        if read_fd in self._read_set:
            i = self._read_set.index(read_fd)
            del self._read_set[i]
            del self._read_handlers[read_fd]
        
    def ignore_write(self, write_fd):
        """
        subsequent write-available events on write_fd ignored.
        """
        if write_fd in self._write_set:
            i = self._write_set.index(write_fd)
            del self._write_set[i]
            del self._write_handlers[write_fd]

    #
    # Timeouts
    #        
    def add_timeout(self, timeout_sec, timeout_handler, run_once=True):
        def cmp_Timeout(x,y):
            if x.time_fire_sec > y.time_fire_sec:
                return 1
            elif x.time_fire_sec < y.time_fire_sec:
                return -1
            else:
                return 0            
        
        time_fire = time() + timeout_sec     
        
        self._timeouts.append(self._Timeout(time_fire, timeout_sec, timeout_handler, run_once))
        self._timeouts.sort(cmp=cmp_Timeout)

    def _handle_timeout(self, time_now):
        """
        runs the handler for _timeouts[0]
        """
        x = self._timeouts[0]
        del self._timeouts[0]
        
        handler = x.timeout_handler
        handler(time_now)

        if not x.run_once:
            self.add_timeout(x.timeout_sec, x.timeout_handler, False)

    def _get_timeouts(self):
        return self._timeouts[:]
    
    #
    # Select loop
    #
    def stop(self):
        self._running = False

    def is_running(self):
        return self._running

    def run(self):
        self._running = True

        time_now = time()
        while self._running:
            # select a timeout for the select call
            select_timeout = None
            if self._timeouts:
                select_timeout = max(0.001,
                                     self._timeouts[0].time_fire_sec - time_now)

            # run the select loop
            rlist,wlist,xlist = select(self._read_set, self._write_set, self._x_set,
                                       select_timeout)
            time_now = time()

            #print "select done: r:%d, w:%d, x:%d" % (len(rlist), len(wlist), len(xlist))
            #print "       sets: r:%d, w:%d, x:%d" % (len(self._read_set), len(self._write_set), len(self._x_set))
            
            # dispatch events
            try:
                if rlist:                 # read-available
                    for fd in rlist:
                        if fd in self._read_handlers:
                            handler = self._read_handlers[fd]
                            handler(fd, time_now)
                                

                if wlist:                 # write-available
                    for fd in wlist:
                        if fd in self._write_handlers:
                            handler = self._write_handlers[fd]
                            handler(fd, time_now)

                # run any expired timeouts
                if self._timeouts and time_now >= self._timeouts[0].time_fire_sec: 
                    while time_now >= self._timeouts[0].time_fire_sec:
                        self._handle_timeout(time_now)
            except (StandardError, Warning):
                einfo = format_exc()
                stderr.write("Exception while dispatching events.\n" + einfo)
                stderr.flush()
                

def wrap(function, *outer_args):
    """
    Returns the closure of the given function with
    the given arguments.

    EXAMPLE:
    
    x = wrap(a, 1, 2)
    x(3, 4)
    x(5, 6)

    this is essentially equivalent to:

    a(1, 2, 3, 4)  # x(3, 4)
    a(1, 2, 5, 6)  # x(5, 6)
    """
    def wrapped_function(*args):
            args = outer_args + args
            function(*args)
    return wrapped_function      

##def hello(): print "hello"
##def bye(): print "bye"
##def ptime():
##    t = time()
##    t2 = int(t) % 1000 + (t - int(t))
##    print "time:", t, "...", t2
##    
##s = SelectLoop()
##
##s.add_timeout(1, ptime, False)
##s.add_timeout(0.5, hello, False)
##s.run()
