#!/usr/bin/python
import sys
import struct
import time

import pymedia.audio.sound as sound
import speex
import asyncio
import horde_api

import stats
mean = stats.lmean
stdev = stats.lstdev

# usage:  ./fileRecv.py > debug.log
#     no command line parameters, but redirect standard out
#     to avoid printing out debug info.

#
# global constants
#
HORDE_SESSION_NAME = "audio"
HORDE_STREAM_NAME = "audio_stream"

AUDIO_SAMPLE_RATE = 8000 # sampling for sent audio
AUDIO_CHANNELS = 1        # number of channels in sent audio
AUDIO_FORMAT = sound.AFMT_S16_LE

SPEEX_QUALITY_LEVEL = 5  # encoding quality: 0 is worst, 8 is best

#
# Freq		:	samples per adu
# 8000		:	170
# 11025		:	235
# 16000		:	341
# 22050		:	470
# 44100		:	940
#
# The best ones for us are 8000 and 22050

SAMPLES_PER_ADU = 170	# will determine it from micSend.py


PLAYBACK_TICK_PERIOD = (((1.0 / AUDIO_SAMPLE_RATE) / AUDIO_CHANNELS) * SAMPLES_PER_ADU) * .9

#
# miscellaneous
#
start_time = time.time()

def debug_timestamp():
    t = time.time() - start_time
    t *= 1000
    return "%8d\t" % int(t)

#
# playback buffer
#
class PlaybackBuffer:
    def __init__(self):
        self.adus = {}
        self.last_popped = None
        
    def __len__(self): return len(self.adus)        

    def add(self, adu_id, data):
        """adds an adu to this playback buffer."""
        self.adus[adu_id] = data

    def pop(self):
        """
        pops the next in-order adu from the buffer and returns
        the tuple (adu_id, data) or (None,None). may also pop other adus
        from the buffer if they were received too out of order.
        """
        if len(self.adus):
            # pop an adu
            adu_id = min(self.adus.keys())
            data = self.adus[adu_id]
            del self.adus[adu_id]

            # should we return this, or is it too old?
            if self.last_popped and adu_id < self.last_popped:
                return self.pop() # too old, find another one
            else:
                self.last_popped = adu_id
                return (adu_id, data)
        else:
            return (None, None) # nothing in the buffer

#
# buffering algorithms
#
class BufferingAlgorithm: # interface used by implementations
    def new_data(self, new_sample_count):
        """called after new samples are added to the playback buffer."""
        pass
    
    def new_rtt(self, rtt):
        """called when a new packet rtt is available."""
        pass
            
    def empty_buffer_at_playback(self, time_now):
        """called when a playback tick finds an empty playback buffer"""
        pass

    
class BasicBufferingAlgorithm(BufferingAlgorithm):
    STATE_INIT = 0
    STATE_BUFFERING = 1
    STATE_PRE_PLAYBACK = 2
    STATE_PLAYBACK = 3
    
    MIN_RTT_SAMPLE_COUNT = 20
    MAX_EMPTY_COUNT_PER_SEC = 5
    MAX_BUF_SIZE = 20
    
    def __init__(self):
        # fields
        self.last_empty = None
        self.empty_count = 0
        self.delay_samples = []
        self.state = self.STATE_INIT

        # enter 'buffering' state
        self.enter_state_buffering()

    # local helpers
    def enter_state_buffering(self):
        global playback_enabled
        #self.delay_samples = []
        playback_enabled = False        
        self.state = self.STATE_BUFFERING

    def enter_state_playback(self):
        global playback_enabled
        #self.delay_samples = []        
        self.last_empty = None
        self.empty_count = 0
        playback_enabled = True       
        self.state = self.STATE_PLAYBACK
        
    def enter_state_pre_playback(self, delay_msec):
        """
        Enters 'pre-playback' state, and moves
        on to playback state after delay_msec.
        """
        global io_loop
        self.state = self.STATE_PRE_PLAYBACK

        if (delay_msec > 0): # enter playback after some delay
            io_loop.add_timeout(delay_msec / 1000.0,
                                asyncio.wrap(BasicBufferingAlgorithm.timer_start_playback, self))
        else: # enter playback immediately
            self.enter_state_playback()

    def timer_start_playback(self, time_now):
        self.enter_state_playback()

    # algorithm interface
    def new_data(self, new_sample_count):
        """called after new samples are added to the playback buffer."""
        if self.state == self.STATE_BUFFERING:
            if len(playback_buffer) > self.MAX_BUF_SIZE:
                self.enter_state_pre_playback(0)
    
    def new_rtt(self, rtt):
        """called when a new packet rtt is available."""
        one_way_delay = rtt / 2
        self.delay_samples.append(one_way_delay)

        if len(self.delay_samples) > self.MIN_RTT_SAMPLE_COUNT:
            if self.state == self.STATE_PLAYBACK:
                del self.delay_samples[0]  # don't want list to grow too big
            elif self.state == self.STATE_BUFFERING:
                # enough samples to estimate delay before playback
                delay_mean = mean(self.delay_samples)
                delay_sdev = stdev(self.delay_samples)
                
                playout_delay = delay_mean + 4 * delay_sdev

                actual_delay = playout_delay - \
                               len(playback_buffer) * SAMPLES_PER_ADU * ((1.0 / AUDIO_SAMPLE_RATE) / AUDIO_CHANNELS)

                print "new_rtt\t%.1f\t%.1f\t%.1f\t%.1f" % (delay_mean, delay_sdev, playout_delay, actual_delay)

                self.enter_state_pre_playback(max(0, actual_delay))
            
    def empty_buffer_at_playback(self, time_now):
        """called when a playback tick finds an empty playback buffer"""

        if self.last_empty:
            if int(time_now) == int(self.last_empty):
                self.empty_count += 1
                if self.empty_count > self.MAX_EMPTY_COUNT_PER_SEC:
                    self.enter_state_buffering()  # need to go back to buffering...
            else: # bit of a hack...
                self.empty_count = 0            
        
        self.last_empty = time_now

#
# initialize global variables
#

# audio related
speex_decoder = speex.new(SPEEX_QUALITY_LEVEL)
audio_out = sound.Output(AUDIO_SAMPLE_RATE, AUDIO_CHANNELS, AUDIO_FORMAT)

# i/o loop
io_loop = asyncio.SelectLoop()

# playback buffer related
playback_enabled = True
buffer_algorithm = BasicBufferingAlgorithm()
playback_buffer = PlaybackBuffer()

#
# horde related functions
#
def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    print "statistics: recv=%d (%s), nchannels=%d"\
                     % (total_recv, rate_to_kbps(total_recv), nchannels)

def show_status():
    print "len(buf): %d, %d   init_over: %s  \r" % (len(buff), empty_tick_count, init_over)

def session_cb(x):
    if (x == horde_api.EVENT_SESSION_CLOSED):
        print "session closed. exiting."
        horde_api.session_close(HORDE_SESSION_NAME)
        sys.exit(0)
        
def recv_cb(stream_handle, adu_id, totlen, offset, data):  # assumes adu's aren't fragmented
    rtt = struct.unpack("h", data[:2])[0]
    audio_data = data[2:]

    playback_buffer.add(adu_id, audio_data)
    buffer_algorithm.new_data(len(audio_data) / 2)

    print debug_timestamp() + "\tRECV\t%d\t%d\t%d" % (adu_id, totlen, rtt)

    if (rtt > 0):
        buffer_algorithm.new_rtt(rtt)
#
# I/O loop callbacks
#
def timer_playback_tick(time_now):
    """
    Called when it is time to play out another frame from
    the playback buffer.
    """
    if playback_enabled:
        # pop decode and playback
        adu_id,data = playback_buffer.pop()
        if data:
            decoded = speex_decoder.decode(data, True)
            audio_out.play(decoded)
            n = len(playback_buffer)
            print "playback buffer: ", n
        else:
            buffer_algorithm.empty_buffer_at_playback(time_now)        

def read_avail_horde_ippc(fd, time_now):
    """
    Called when one or more IPPC messages are available
    on horde's IPPC socket.
    """
    n = 1
    while (n > 0):
        n = horde_api.event_dispatch(0)
        
    if n < 0: # error (daemon probably died)
        print "\nhorde IPPC read error: exiting program."
        sys.exit(1)
               
#
# main loop
#
if (horde_api.session_accept(HORDE_SESSION_NAME)):
        print "client connected."
        h = horde_api.stream_accept(HORDE_STREAM_NAME)
        if (h > 0):
            hstream = h
            print "\tstream '%s' connected, uid:%d" % (HORDE_STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)

            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)
            
            # enter the I/O loop
            fd = horde_api.get_ippc_fd()
            if (fd != -1):
                io_loop.want_read(fd, read_avail_horde_ippc)
                io_loop.add_timeout(PLAYBACK_TICK_PERIOD, timer_playback_tick, run_once=False)
                io_loop.run()
