"""
Extra Horde-API related functionality.

author: Asfandyar
date:   august 23rd 2006
"""
import horde_api
import sys

class ReliableOrderedStream:
    """
    Higher-level stream class that abstracts away much of the horde API
    to provide a TCP-like reliable in-order stream.

    ADU's are always delivered in-order and the receive callback for
    this class never sees fragmentation.    
    """
    #
    # public functions
    #
    def __init__(self, hstream, recv_handler, close_handler=None, debug=False):
        """
        Create a reliable in-order-receive ADU stream using the given stream handle.    

        PARAM recv_handler: data receive callback.
            A function of the following form:
               recv_handler(stream_handle, int adu_id, string adu_data)
            ADUs will be delivered in order to this function when
            all fragments have been received.

        PARAM close_handler: stream-close event handler.
            A function of the following form:
               close_handler(stream_handle)

        NOTES:
        * The first ADU MUST be assigned ID 0 and the identifiers MUST
          increase by one progressively (ie. 0, 1, 2, 3, 4...)
        * You'll need a horde event dispatch loop outside this somewhere.
        * Both sides of the connection need to use this class otherwise
          it won't work.        
        * hstream should NOT be used in horde API calls outside
          this class after this constructor is called.
        """
        self._hstream = hstream

        self._send_buf = {}

        self._recv_buf = {}
        self._next_recv_id = 0

        self._debug = debug
        
        self._close_handler = close_handler
        self._recv_handler = recv_handler        
        
        horde_api.stream_set_recv_callback(hstream, self._recv_callback)
        horde_api.stream_set_event_callback(hstream, self._event_callback)

    def send(self, adu_id, adu_data, allow_fragmentation=1, send_priority=0):
        """Send an ADU. May block."""
        horde_api.stream_send(self._hstream, adu_id, adu_data, priority=send_priority,
                              fragment_allowed=allow_fragmentation)
        self._send_buf[adu_id] = (adu_data, [], send_priority, allow_fragmentation)
        
    #
    # private helpers
    #
    def __complete_adu(self, n, frag_list):
        """tests if the given acks cover all the data."""
        # probably more complex test than needed...
        present = [False] * n
        for x in frag_list: # ack is (offset, len) fragment info
            offset = x[0]
            flen = x[1]
            
            present[offset:offset+flen] = [True] * flen
        
        complete = reduce(lambda x,y: x and y, present)

        return complete

    def __reconstruct_recvd_adu(self, totlen, recv_event_list):
        """tries to reconstruct the adu. returns the complete, reconstructed
           ADU or None if the ADU is not completely defined in recv_event_list"""
        el = map(lambda x: (x[0], len(x[1])), recv_event_list)
        if self.__complete_adu(totlen, el):
            d = [' '] * totlen
            for x in recv_event_list:
                offs = x[0]
                fdata = x[1]
                d[offs:offs+len(fdata)] = fdata
            return "".join(d)
        else:
            return None

    #
    # event handlers
    #
    def _event_callback(self, event, hstream, adu_id, fstart, flen, rtt):
        if (event == horde_api.EVENT_ADU_FRAG_ACK):
            if self._debug: sys.stderr.write("stream[%3d] ACK: adu=%d, offs=%d, len=%d\n" % (hstream, adu_id, fstart, flen)); 
            if (adu_id in self._send_buf): # remember ACK
                e = self._send_buf[adu_id]
                data = e[0]
                acks = e[1]
                acks.append( (fstart, flen) )
                if self.__complete_adu(len(data), acks):
                    del self._send_buf[adu_id]
            elif self._debug: # this may be a bug...
                sys.stderr.write("oops: this line should probably not be reached in ReliableOrderedStream._event_callback(...).\n");
        elif (event == horde_api.EVENT_ADU_FRAG_LOSS):
            if self._debug: sys.stderr.write("stream[%3d] LOSS: adu=%d, offs=%d, len=%d\n" % (hstream, adu_id, fstart, flen));                
            if (adu_id in self._send_buf): # resend data
                e = self._send_buf[adu_id]
                data = e[0]
                acks = e[1]
                pri = e[2]
                frag = e[3]
                horde_api.stream_send_fragment(hstream, adu_id, fstart, flen, data[fstart:fstart+flen],
                                               priority=pri, fragment_allowed=frag)
            elif self._debug: # definitely a bug
                sys.stderr.write("bug: this line should definitely not be reached in ReliableOrderedStream._event_callback(...).\n");
        elif (event == horde_api.EVENT_STREAM_CLOSED):
            if self._debug: sys.stderr.write("stream[%3d] CLOSED.\n" % hstream);         
            if (self._close_handler):
                f = self._close_handler
                f(hstream)

                
    def _recv_callback(self, hstream, adu_id, totlen, frag_start, frag_data):
        if self._debug: sys.stderr.write("stream[%3d] RECV: adu=%d, offs=%d, len=%d, next_recv=%d\n" %\
                                         (hstream, adu_id, frag_start, len(frag_data), self._next_recv_id));
        
        if not (adu_id in self._recv_buf): self._recv_buf[adu_id] = []
        self._recv_buf[adu_id].append( (frag_start, frag_data) )

        if self._next_recv_id in self._recv_buf:
            recv_cb = self._recv_handler
            while self._next_recv_id in self._recv_buf:                
                # try to reconstruct the adu and invoke the callback if complete ADU
                adu = self.__reconstruct_recvd_adu(totlen, self._recv_buf[self._next_recv_id])
                if not adu: break
                if self._debug: sys.stderr.write("stream[%3d] RECV: adu %d complete and in-order.\n" % (hstream, self._next_recv_id));
                recv_cb(hstream, self._next_recv_id, adu)
                del self._recv_buf[self._next_recv_id]
                self._next_recv_id += 1
                    
        if self._debug:
            k = self._recv_buf.keys()
            k.sort()
            sys.stderr.write("recv buffer: %s\n" % str(k))
        
        
    

