#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <vector>
#include <iostream>

#include <nmstl/ntime>

#include "properties.h"
#include "daemon.h"
#include "net.h"
#include "ippc/ippc_server.h"

using namespace nmstl;
using namespace std;
using namespace horde;

//
// variables
//
properties horde_properties;

//
// remembering old session requests
//
struct old_session_request_t {
  string session_id;
  inet_address peer_address;
  ntime timestamp;

  old_session_request_t(string& id, inet_address& addr, ntime& t) 
    : session_id(id), peer_address(addr), timestamp(t) { }
};
vector<old_session_request_t> received_connect_requests; ///< list of received accept requests

// requests should be discarded after 3 seconds.
#define REQUEST_TIMEOUT 3000

// retreives and removes an old entry for the named session and returns true, or
// returns false and doesn't modify received_connect_requests.
bool get_old_session_request(string session_id, inet_address* peer_address, ntime* timestamp) {
  // If there is an old connect request for this session, 
  // remove it from the vector
  for (vector<old_session_request_t>::iterator iter = received_connect_requests.begin();
       iter != received_connect_requests.end(); iter++) {
    if (iter->session_id == session_id) {
      if (peer_address != NULL) *peer_address = iter->peer_address;
      if (timestamp != NULL) *timestamp = iter->timestamp;
      received_connect_requests.erase(iter);
      return true;
    }
  }
  return false;
}

// finds and removes an old session request for the given id. if it
// finds a matching request and that request hasn't timed out, this
// returns true and the associated peer_address; otherwise it returns false.
bool get_old_session_request(string session_id, inet_address& peer_address) {
  ntime t;
  if (get_old_session_request(session_id, &peer_address, &t)) {
    if (ntime::now_plus_msecs(REQUEST_TIMEOUT) >= t) {
      return true;
    }
  }

  return false;
}

void remember_session_request(string session_name, inet_address peer_address) {
  // remove an older entry for this session_id, if it exists...
  get_old_session_request(session_name, NULL, NULL);

  // remember
  ntime now = ntime::now();
  received_connect_requests.push_back(old_session_request_t(session_name, peer_address, now));
  LOG_EVENT("session") << "remembering connect request for session: '" << session_name << "'";
}

//
// basic stuff
//
string get_logfilename() {
  // get time string...
  char buf[128];
  ntime now = ntime::now();
  time_t sec = now.to_secs();
  tm t;
  if (!gmtime_r(&sec, &t)) return "";
  
  size_t ret = strftime(buf, sizeof buf, "%d.%b.%Y.%H.%M.%S", &t);
  if (ret)
    return session_daemon_out_redirect + buf;
  else
    return "";
}

// runs a command and then waits till it is completed.
int spawn_and_wait(char *command, char* argv[], char* envp[], const char* logfilename) {
  pid_t child = fork();
  
  if (child == 0) { 
    // (in child)    
    int id;
    if (logfilename != NULL) {
      id = open (logfilename, O_WRONLY | O_CREAT);
      if (id != -1) {
	close(1);
	close(2);
	dup(id);
	dup(id);
      }
    }

    int ret = execve(command, argv, envp); // run the command.

    if ((logfilename != NULL) && (id != -1)) {
      close(1);
      close(2);
      close(id);
    }
  
    LOG_EVENT_DEBUG("error") << "exec() failed: " << strerror(errno) << ", err=" << errno;
    ::exit(0);
  } else if (child > 0) { 
    // (in parent)
    // wait for child to exit...
    int status=0;
    pid_t x = wait(&status);
    return status;
  } else { 
    // (in parent)
    LOG_EVENT_DEBUG("error") << "fork() failed: " << strerror(errno) << ", err=" << errno;
    return -1;
  }
}

// cleans up zombie processes left around from the fork in spawn... (10 second period)
#define SPAWN_CLEANUP_PERIOD 10000
void spawn_cleanup_timer_fire() {
  // clean up a child
  waitpid(-1, NULL, WNOHANG | WUNTRACED);

  // run this again in a while...
  (new callback_timer(io_loop, spawn_cleanup_timer_fire))->arm(ntime::now_plus_msecs(SPAWN_CLEANUP_PERIOD));
}

void horde::init_zombie_killer() {
  spawn_cleanup_timer_fire();
}

// recursively deletes all children of a directory
// and eventually deletes that directory.
void 
horde::remove_directory(const string& dir_path) {
  //
  // make sure directory is empty...
  //
  struct dirent **namelist;
  int n;
  
  n = scandir(dir_path.c_str(), &namelist, 0, alphasort);
  if (n < 0) {
    LOG_EVENT_DEBUG("error") << "remove_directory failed for \'" << dir_path 
			     << "\', reason:" << strerror(errno);
    return;
  }
  else {
    while(n--) {
      if ((strcmp(".", namelist[n]->d_name) == 0) ||
	  (strcmp("..", namelist[n]->d_name) == 0))
	continue;

      string fpath = dir_path + '/' + namelist[n]->d_name;

      //printf("%s\n", namelist[n]->d_name);
      if (namelist[n]->d_type == DT_DIR) {
	remove_directory(fpath);
      } else {
	unlink(fpath.c_str());
      }

      free(namelist[n]);
    }
    free(namelist);
  }
  
  //
  // remove the directory
  //
  if (rmdir(dir_path.c_str()) == 0) {
    LOG_EVENT_DEBUG("fileio") << "removed directory: " << dir_path << " [" << (n+1) << " children]";
  }
}

//
// spawning a session daemon
//
char** split_command_line(const string& command) {
  // parse string
  vector<string> elements;
  int n = 0;
  string current="";
  bool in_space = true;
  for (int i = 0; i < command.size(); i++) {
    if ((command[i] == ' ') && (in_space == false)) { // first space in a space-block
      in_space = true;
      elements.push_back(current);
      current = "";
    } else {
      current += command[i];

      if (in_space == true) { // space-block just ended
	in_space = false;
      }
    }
  }
  if ((in_space == false) && (current != "")) {
    elements.push_back(current);
  }

  // convert to a char[]
  n = elements.size();
  char** argv = (char**)calloc(sizeof(char*), n+1);
  for (int i = 0; i < n; i++) {
    size_t sz = elements[i].size()+1;
    argv[i] = new char[sz];
    strncpy(argv[i], elements[i].c_str(), sz);    
  }
  argv[n] = NULL;

  return argv;
}

pid_t spawn_session_daemon(string session_name, string peer_addr, bool accepting, int pipe_fd) {
  pid_t child = fork();
  if (child == 0) { // (in child)
    string session_name_root = namespace_root + "/sessions/" + session_name;    

    // create directory
    mkdir(session_name_root.c_str(), 0777);

    // pipe file descriptor
    char pipefdstr[6];
    snprintf(pipefdstr, 6, "%d", pipe_fd);
    
    // spawn session daemon and wait for it to end...

    string command;
    if (session_daemon_wrapper != "") { // no wrapper
      command += session_daemon_wrapper + " ";
    }
    command += session_daemon_command;
    command += " ";
    command += properties_filename;
    command += " ";
    if (accepting) command += "ACCEPT"; else command += "CONNECT";    
    command += " ";
    command += session_name;
    command += " ";
    command += peer_addr;
    command += " ";
    command += pipefdstr;

    char** argv = split_command_line(command);
    char* com = argv[0];

    string logname;
    const char* logfilename = NULL;
    if (session_daemon_out_redirect != "") {
      logname = get_logfilename();
      logfilename = logname.c_str();
    }

    char* envp[] = {(char*)NULL};

    LOG_EVENT("spawn") << "spawning session daemon: '" << argv[0] << "' '" <<  argv[1] << "' '"
		       << argv[2] << "' '" << argv[3] << "' '" << argv[4] << "' '" <<  argv[5] 
		       << "'... log: " << logfilename;

    int stat = spawn_and_wait(com, argv, envp, logfilename);

    // remove directory
    remove_directory(session_name_root);

    LOG_EVENT_DEBUG("spawn") << "cleaned up after session daemon: " << session_name << ", status:" << stat; 

    // exit
    ::exit(0);
  }

  return child; // (in parent)
}

//
// properties stuff
//  
bool 
horde::load_properties_file(const string& filename) {
  int err = horde_properties.read_from_file(filename);
  
  return !(err < 0);
}
  
string 
horde::get_property(string property, string default_value) {
  return horde_properties.get(property, default_value);
}
  
void 
horde::set_property(string property, string value) {
  horde_properties.set(property, value);
}

//
// session stuff
//
struct session_t { // a session, ready to accept 
  string name;
  ippc_handler* source;

  session_t(string name, ippc_handler* source) :
    name(name), source(source) { }
};

// contains the sessions that are in either in
// the READY_TO_ACCEPT, TRYING_TO_CONNECT, or CLOSED states.
vector<session_t> sessions;

vector<session_t>::iterator lookup_session(string session_name) {
  vector<session_t>::iterator iter = sessions.begin();
  for (; iter != sessions.end(); iter++) {
    if (iter->name == session_name) break;
  }

  return iter;
}

void cleanup_session_info(vector<session_t>::iterator& iter) {
  LOG_EVENT_DEBUG("session") << "cleaning up local state: " << iter->name;

  sessions.erase(iter);
}

bool is_valid_name_string(const char *str, size_t len) {
  for (int i = 0; i < len; i++) {
    char ch = str[i];
    if (!(((ch >= 'A') && (ch <= 'Z')) ||
	  ((ch >= 'a') && (ch <= 'z')) ||
	  ((ch >= '0') && (ch <= '9')) ||
	  (ch == '_'))) 
      return false; // not alph/num/underscore
  }

  return true;
}

bool is_session_daemon_running(const char* session_name) {
  // need to check namespace for active sessions with this name
  string dir = namespace_root + "/sessions/" + string(session_name);
  
  if (chdir(dir.c_str()) == 0) {
    return true; // directory exists, so session name is in use.
  } else {
    if (errno == ENOENT) {
      return false; // name seems okay...
    } else {
      LOG_EVENT_DEBUG("debug") << "Unexpected errno: " << errno << "[" << strerror(errno) << "]";
      return true; // something else went wrong...
    }
  }      
}

bool is_valid_session_name(const char* session_name) {
  size_t name_length = strlen(session_name);

  // allowed name rules:
  // 1. must be alphanumeric+underscores < 32 chars.
  // 2. must not be an entry in sessions with this name
  // 3. no such sub-directory in: NAMESPACE_ROOT + '/sessions'

  if (!is_valid_name_string(session_name, name_length)) {
    // bad name string.
    return false; 
  } else if (lookup_session(string(session_name)) != sessions.end()) {
    // an entry in sessions with this name.
    return false;
  } else {
    return !is_session_daemon_running(session_name);
  }
}
 
void 
horde::session_connect(ippc_handler* request_source,
		       const char* session_name, const char* peer_address) {

  if (is_valid_session_name(session_name)) {
    // ... fork off ...
    pid_t child = spawn_session_daemon(string(session_name), string(peer_address), 
				       false, request_source->get_fd());    
    if (child > 0) {
      request_source->close(); // other process is taking over this pipe...
    } else {
      // send an exception to the owner
      deliver_exception_internal(request_source, "Internal horde error. Session closed.");

      // log the problem.
      LOG_EVENT_DEBUG("session") << "ERROR: could not spawn session daemon process: " 
				 << strerror(errno) << ", errno=" << errno;
    }
  } else {
    string msg = "'" + string(session_name) + "' is not a valid session ID.";
    deliver_exception_bad_session_id(request_source, msg.c_str());
  }
}

void 
horde::session_accept(ippc_handler* request_source, const char* session_name) {
  string session_id(session_name);
  if (is_valid_session_name(session_name)) {    
    // create an entry for the session so we react properly to connect requests.
    sessions.push_back(session_t(session_id, request_source));

    LOG_EVENT("session") << "ready to accept: \'" << session_id << "\'";



    // If there is an old connect request for this session, 
    // handle it now.
    inet_address peer;
    if (get_old_session_request(session_id, peer)) {
      handle_recv_connect_request(peer, session_id);
    }
  } else {    
    string msg = "'" + session_id + "' is not a valid session ID.";
    deliver_exception_bad_session_id(request_source, msg.c_str());
  }
}

void 
horde::session_close(ippc_handler* ioh, const char* session_id) {

  for (vector<session_t>::iterator iter = sessions.begin();
       iter != sessions.end();
       iter++) {
    if (iter->name == session_id) {
      // found a session with the given name

      ippc_handler* owner = iter->source;
      cleanup_session_info(iter);
      deliver_event_closed(owner, session_id);
      break;
    }
  }
}

//
// functions called in response to network events
//

void 
horde::handle_recv_connect_request(inet_address peer_address, string session_name) {
  LOG_EVENT("session") << "recv connect REQUEST [name: '" << session_name 
		       << "', from: '" << peer_address.as_string() << "'";

  vector<session_t>::iterator iter = lookup_session(session_name);
  if (iter != sessions.end()) {
    pid_t child = spawn_session_daemon(session_name, peer_address.as_string(), 
				       true, iter->source->get_fd());
    
    if (child > 0) {
      iter->source->close(false); // other process is taking over this pipe...
    } else {
      // send an exceptio to owner
      deliver_exception_internal(iter->source, "Internal horde error. Session closed.");

      // log the problem.
      LOG_EVENT_DEBUG("session") << "ERROR: could not spawn session daemon process: " 
				 << strerror(errno) << ", errno=" << errno;
    }
    
    // cleanup
    cleanup_session_info(iter);
  } else {
    // remember this session request if a daemon isn't already running...
    if (!is_session_daemon_running(session_name.c_str()))
      remember_session_request(session_name, peer_address);
  }
}

void 
horde::handle_recv_connect_ack(inet_address peer_address, string session_name) {
  // ... this daemon should never get an ack
}

void
horde::handle_recv_connect_nack(inet_address peer_address, string session_name) {
  // ... no more nacks...
}

// clean up any sessions this owner owns...
void
horde::remove_sessions_owned_by(ippc_handler* owner) {
  vector<session_t>::iterator iter = sessions.begin();
  while (iter != sessions.end()) {
    // find an entry to delete
    for (; iter != sessions.end(); iter++) {  
      if (iter->source == owner) break;    
    }
    
    // if last loop found an entry to delete, get rid of it.
    if (iter != sessions.end()) {
      cleanup_session_info(iter);

      // reset iterator (maybe other items to remove)
      iter = sessions.begin();   
    }        
  }
}
