// -*- mode: C++; -*-

/** @file daemon.h
 *
 * Shared variables and functions in the horde dispatch daemon.
 */
#ifndef __HORDE_DAEMON_DISPATCH_H__
#define __HORDE_DAEMON_DISPATCH_H__

#include <dirent.h>
#include <unistd.h>

#include <nmstl/net>
#include <nmstl/ioevent>

#include <string>
#include "horde_log.h"
#include "horde_namespace.h"

using namespace std;
using namespace nmstl;
using namespace horde;

extern io_event_loop io_loop;             ///< Daemon's only I/O event loop
extern string namespace_root;             ///< Root filename for the IPPC namespace
extern string properties_filename;        ///< Filename from where the properties are loaded.
extern int primary_port;                  ///< Primary network port (UDP)

/// Name of session daemon executable.
extern string session_daemon_command;     

/// A wrapper command ("/usr/bin/valgrind") or "" for no wrapper.
extern string session_daemon_wrapper;     

/// A filename to redirect the session daemon's stdout+stderr (file will be truncated) or "" for no redirection.
extern string session_daemon_out_redirect; 

HORDE_NAMESPACE_BEGIN;

//
// misc stuff
//

/**
 * Remove a directory and any contained files and sub-directories.
 */
void remove_directory(const string& dir_path);

//
// ippc stuff
//

class ippc_handler;

/**
 * Initialize the system call server. Starts an IPPC server bound to
 * the given path.
 */
void init_syscall(io_event_loop& ioloop, string path);

/**
 * Cleanly shut down the system call server. Clean up any file name
 * additions (e.g., named pipes).
 */
void shutdown_syscall();

//
// properties
//

/**
 * Load properties from a file.
 */
bool load_properties_file(const string& filename);

/**
 * Get the value of a property. Both name and value are case-sensitive.
 * @return value of named property, or \a default_value.
 */
string get_property(string property, string default_value);

/**
 * Set the value of a property. Both name and value are case-sensitive.
 */
void set_property(string property, string value);

//
// session stuff
//

/**
 * Attempt a session connect. If the session ID is not valid, an
 * exception is delivered to the specified IPPC client. 
 * 
 * Otherwise, a session daemon process is spawned and the IPPC client
 * is transferred to that process.  The session daemon will send
 * network requests and try to establish a connection.
 *
 * @param peer_address A valid internet address (e.g., \c "128.30.76.73",
 *                     or \c "slice.csail.mit.edu").
 */ 
void session_connect(ippc_handler* ioh, const char* session_id, const char* peer_address);

/**
 * Attempt a session accept. If the session ID is valid, a session is created
 * and put into the waiting state, until a remote client tries to connect. 
 * An invalid ID results in an exception being delivered to the specified
 * IPPC client.
 *
 * When a remote client tries to connect to the given session, a session daemon
 * process is spawned and the IPPC client is transferred to that process.
 */ 
void session_accept(ippc_handler* ioh, const char* session_id);

/**
 * Close a session if it is waiting. If a session is closed, a 'session closed'
 * event will be delivered to the specified IPPC client; if no session is closed,
 * nothing will happen.
 */ 
void session_close(ippc_handler* ioh, const char* session_id);

/** 
 * Remove any waiting sessions that are owned by the given IPPC client. 
 * Any sessions that are already connected will not be closed.
 */
void remove_sessions_owned_by(ippc_handler* ioh);

//
// hooks from/to network code
//

/// Handle the reception of a connection request on the network.
void handle_recv_connect_request(inet_address peer_address, string session_name);

/// Handle a connect ACK packet. @todo remove
void handle_recv_connect_ack(inet_address peer_address, string session_name);

/// Handle a connect NACK packet. @todo remove
void handle_recv_connect_nack(inet_address peer_address, string session_name);

class interface_pool_manager;

/// Get the network channel pool manager.
interface_pool_manager* get_poolman();

//
// deliver events
//

/// Deliver an 'internal exception' notification to the given IPPC client.
void deliver_exception_internal(ippc_handler* ioh, const char* msg);

/// Deliver a 'bad session id exception' notification to the given IPPC client.
void deliver_exception_bad_session_id(ippc_handler* ioh, const char* msg);

/// Deliver a 'session closed' event to the given IPPC client.
void deliver_event_closed(ippc_handler* ioh, const char* session_id);

//
// miscellaneous
//

/// initialize a periodic process to kill off zombie children...
void init_zombie_killer();

HORDE_NAMESPACE_END;

#endif
