//
// functions to handle IPPC calls.
//
#include <nmstl/ioevent>
#include <nmstl/net>

#include <unistd.h>

#include "ippc/ippc_server.h"
#include "ippc/syscall.h"
#include "horde_log.h"

#include "daemon.h"

using namespace horde;
using namespace nmstl;

ippc_server* server;

//
// Initialization
//

void recv_handler(ippc_handler* ioh, int type,char* msg, size_t len);
void close_handler(ippc_handler* ioh);

// clean up IPPC namespace (destroy all sessions)
void cleanup_namespace() {
  string session_dir = namespace_root + "/sessions";
  remove_directory(session_dir);
  mkdir(session_dir.c_str(), 0777);
}

// initialize and start syscall server bound to the given path
void horde::init_syscall(io_event_loop& ioloop, string path) {
  cleanup_namespace(); // TODO: test if another server is running, or if...  
  server = new ippc_server(io_loop, path, recv_handler, close_handler);
}

void horde::shutdown_syscall() {
  string path = server->get_path();
  server->stop_server(false);
  unlink(path.c_str());
}

//
// IPPC event handlers
//

void send_exception(ippc_handler* ioh, ippc_exception_code excode, const char* msg) {
  ippcmsg_exception ex;
  bzero(&ex, sizeof ex);
  
  ex.code = excode;
  strncpy(ex.msg, msg, EXCEPT_MSG_MAX_LENGTH);

  ioh->send_message(IPPC_EXCEPTION, &ex, sizeof ex);
}

// handle an IPPC receive
void recv_handler(ippc_handler* ioh, int type,char* msg, size_t len) {
  if (type == IPPC_SESSION_COMMAND) {
    ippcmsg_session_command* com = (ippcmsg_session_command*)msg;           
    
    // try to execute the command
    if (com->command == ippcmsg_session_command::CONNECT) {
      // log
      LOG_EVENT("ipcc") << "IPPC_SESSION_COMMAND: command=CONNECT, peer=" << com->peer_address
			<< "session_id='" << com->session_id << "'";

      // todo: validate parameters?
      
      // execute command
      inet_address peer(com->peer_address, primary_port);

      session_connect(ioh, com->session_id, peer.as_string().c_str() );
      
    } else if (com->command == ippcmsg_session_command::ACCEPT) {
      // log
      LOG_EVENT("ipcc") << "IPPC_SESSION_COMMAND: command=ACCEPT, session_id='" 
			<< com->session_id << "'";

      // execute command
      session_accept(ioh, com->session_id); // @todo test to make sure valid inputs?
    } else if (com->command == ippcmsg_session_command::CLOSE) {
      // log
      LOG_EVENT("ipcc") << "IPPC_SESSION_COMMAND: command=CLOSE, session_id='" 
			<< com->session_id << "'";

      // execute command
      session_close(ioh, com->session_id);
    } else {
      // log the bad ippc code
      LOG_EVENT("ipcc") << "Bad IPPC_SESSION_COMMAND code: " << (int)com->command;
    
      // reply with an exception
      send_exception(ioh, IPPC_EX_BAD_IPPC, "Bad IPPC_SESSION_COMMAND code.");
    }
    
  } else if (type == IPPC_NOP) {
    // do nothing
    LOG_EVENT_DEBUG("ipcc") << "NOP";
  } else {
    // log the bad ippc code
    LOG_EVENT("ipcc") << "Bad IPPC code: " << type << " with " << len << "-byte message.";
    
    // reply with an exception
    send_exception(ioh, IPPC_EX_BAD_IPPC, "Bad IPPC code: only IPPC_SESSION_COMMAND allowed.");
  }
}

// handle an IPPC close
void close_handler(ippc_handler* ioh) {
  if (ioh == NULL) {
    horde::panic("IPPC server exited unexpectedly.");
  } else {
    remove_sessions_owned_by(ioh);
    LOG_EVENT("ipcc") << "client closed.";
  }
}

// deliver events
void 
horde::deliver_exception_internal(ippc_handler* ioh, const char* msg) {
  send_exception(ioh, IPPC_EX_INTERNAL, msg);
}

void 
horde::deliver_exception_bad_session_id(ippc_handler* ioh, const char* msg) {
  send_exception(ioh, IPPC_EX_BAD_SESSION_ID, msg);
}

void 
horde::deliver_event_closed(ippc_handler* ioh, const char* session_id) {
  ippcmsg_session_event evt;
  bzero(&evt, sizeof evt);

  evt.event = ippcmsg_session_event::CLOSED;
  strncpy(evt.session_id, session_id, SESSION_ID_MAX_LENGTH);  

  ioh->send_message(IPPC_SESSION_EVENT, &evt, sizeof(evt));
}
