#include <iostream>
using namespace std;

#include "coder.h"

typedef unsigned char byte;

//
// helpers
//

void _xor_blocks(byte* dest, vector<byte*>& window, size_t size) {
  //  printf("--| XOR: ");
  memset(dest, 0, size);
  for (vector<byte*>::iterator i = window.begin(); i != window.end(); i++) {    
    byte* src = (*i);
    if (src) { // ignore NULL entries
      //      printf("%d ", *((int*)src));
      for (int j = 0; j < size; j++) {
	dest[j] = dest[j] ^ src[j];
      }
    }
  }
  //  printf("= %d |--\n", *((int*)dest));
}

byte* _copy_block(const byte* block, size_t bsize, size_t size) {
  byte *b = new byte[size];
  memset(b, 0, size);
  memcpy(b, block, bsize);

  return b;
}

void _window_push(const byte* block, size_t bsize, size_t size, vector<byte*>& window) {
  window.push_back( _copy_block(block, bsize, size) );
}

void _window_clear(vector<byte*>& window) {
  for (vector<byte*>::iterator i = window.begin(); i != window.end(); i++) {    
    delete [] (*i);
  }
  window.clear();
}

//
// coder
//
pair<unsigned char*,int> 
BlockCoder::push(const byte *block, int blockid, size_t size) {
  _window_push(block, size, block_size, window);

  if (window.size() == window_size) {
    _xor_blocks(coded, window, block_size);
    _window_clear(window);

    return pair<unsigned char*,int>(coded, blockid+1);
  } else {
    return pair<unsigned char*,int>(NULL, 0);
  }
}

//
// decoder
//
void 
BlockDecoder::_window_map_erase(int id) {
  window_map_t::iterator i = windows.find(id);
  if (i != windows.end()) {
    _window_map_clear_row(i->second);
    windows.erase(i);
  }
}

void 
BlockDecoder::_window_map_clear_row(window_t& vec) {
  for (window_t::iterator j = vec.begin(); j != vec.end(); j++) {
    if (j->first) delete [] j->first;
  }
}

#define MAX_WINDOW_MAP_SIZE 64
void 
BlockDecoder::_window_map_limit_size() {
  if (windows.size() > MAX_WINDOW_MAP_SIZE) {
    window_map_t::iterator i = windows.begin();
    _window_map_clear_row(i->second);
    windows.erase(i);
  }
}

void 
BlockDecoder::push(unsigned char* block, int blockid, size_t bsize) {  
  int j;
  int windowid = blockid / (window_size + 1);
  int woffset = blockid % (window_size + 1);

  pair<unsigned char*, bool> p(_copy_block(block, bsize, block_size), false);

  if (woffset < window_size) { // not a coded block
    decoded_block_ready_cb(p.first, blockid, cb_arg); 
    p.second = true;
  } 

  window_map_t::const_iterator i = windows.find(windowid);
  if (i == windows.end()) {
    windows[windowid] = window_t(window_size+1);
    for (int j = 0; j < (window_size+1); j++) windows[windowid][j].first = NULL;

    _window_map_limit_size();
  }

  windows[windowid][woffset] = p;

  // see if enough entries in current window to decode successfully.
  int ndecoded = 0, npresent=0;
  for (j = 0; j < (window_size+1); j++) {
    if (windows[windowid][j].first) {
      npresent++;
      if (windows[windowid][j].second) ndecoded++;
    }
  }

  if (ndecoded == window_size) {
    // we're done with this window
    _window_map_erase(windowid);
  } else if (npresent == window_size) {
    // invariant: two blocks still undecoded (the coded one, and a missing one)
    //   why: coded one never decoded && decoded ones always decoded &&
    //        npresent == window_size && ndecoded < window_size
    int decodedj=0;
    vector<byte*> blocks;
    for (j = 0; j < window_size; j++) {
      blocks.push_back(windows[windowid][j].first);
      if (!windows[windowid][j].second) decodedj = j;
    }
    blocks.push_back(windows[windowid][window_size].first);

    int blockid = windowid * (window_size+1) + decodedj;

    byte* decoded = new byte[block_size];
    _xor_blocks(decoded, blocks, block_size);

    //    cout << "decodedj=" << decodedj << ", blockid=" << blockid << endl;
    
    decoded_block_ready_cb(decoded, blockid, cb_arg);

    _window_map_erase(windowid);
  }
}

//
// debug code
//
#if 0

typedef double etype;
vector<etype> deco;
vector<int> decid;

void decoded_cb(byte* buf, int id, void* arg) {
  deco.push_back(*(etype*)buf);
  decid.push_back(id);

  //cout << "decoded_cb: " << id << " -> " << *((int*)buf) << endl;
}

template<typename T>
void print_vec(vector<T> v, vector<int>* v2=NULL, const char* pref="\t") {
  for (int i = 0; i < v.size(); i++) {
    cout << pref << i << ":\t" << v[i];
    if (v2) cout << "\t" << (*v2)[i];
    cout << endl;
  }
}

int main() {
  BlockCoder coder(sizeof(etype), 3);
  BlockDecoder decoder(decoded_cb, sizeof(etype), 3);

  pair<byte*, int> p;

  vector<etype> src, code;
  vector<int> ids;
  deco.clear();
  decid.clear();

  // generate source
  for (int i = 0; i < 10; i++) src.push_back(i / 3.125);
  cout << "SRC = " << endl;
  print_vec(src);
  cout << endl;

  // code
  int id = 0;
  for (int i = 0; i < src.size(); i++) {
    p = coder.push((byte*)&(src[i]), id, sizeof(etype));

    code.push_back(src[i]);
    ids.push_back(id);
    
    if (p.first) {
      id = p.second;
      code.push_back(*((etype*)p.first));
      ids.push_back(id);
    }

    id++;
  }
  cout << "CODED = " << endl;
  print_vec(code, &ids);
  cout << endl;

  // drop some
  vector<etype> dropped;
  vector<int> droppedids;
  for (int i = 0; i < code.size(); i++) {
    if (i % 4) {
      dropped.push_back(code[i]);
      droppedids.push_back(ids[i]);
    }
  }
  code = dropped;
  ids = droppedids;
  cout << "DROPPED = " << endl;
  print_vec(code, &ids);
  cout << endl;

  // decode
  for (int i = 0; i < code.size(); i++) {
    decoder.push((byte*)&(code[i]), ids[i], sizeof(etype));
  }
  cout << "DECODED = " << endl;
  print_vec(deco, &decid);
  cout << endl;

  vector<etype> d;
  for (int i = 0; i < 16; i++) d.push_back(0);
  //d.resize(code.size());
  for (int i = 0; i < deco.size(); i++) {
    d[decid[i]] = deco[i];
  }
  cout << "OUT = " << endl;
  print_vec(d);
  cout << endl;
}

#endif
