#define __STDC_LIMIT_MACROS
#include <stdint.h>

#include <set>

#include "scheduler_impl.h"
#include "control_netio.h"
#include "session_daemon.h" // for get_property

#include "net/poolman.h" // statistics

HORDE_NAMESPACE_BEGIN;

static const string DEFAULT_SCHEDULER = "grunt";

static const int SCHEDULE_TX_SLOT_MAX_LOOPREVS = 64;
static const int MIN_FRAGMENT_LEN = 32; // never fragment ADU's smaller than this.

string debugstr(char* data, int sz);

//
// scheduler factory
//
scheduler* create_scheduler(string type_id) {
  if (type_id == "default") type_id = DEFAULT_SCHEDULER;

  if (type_id == "roundrobin") {
    return new scheduler_roundrobin();
  } else if (type_id == "hoard") {
    return new scheduler_hoard(); 
  } else if (type_id == "grunt") {
    return new scheduler_grunt(); 
  } else if (type_id == "firehose") {
    return new scheduler_firehose();
  } else if (type_id == "coder") {
    return new scheduler_coder();
  } else {
    LOG_EVENT("schedule") << "create_scheduler(...) given unknown scheduler type:" << type_id << ", using default type.";
    return new scheduler_roundrobin();
    //return NULL;
  }
}

//
// firehose
//

void
scheduler_firehose::schedule_tx_slot(ptr<tx_slot> slot) {
  size_t n = slot->get_max_size();
  horde_net_packet pkt(n);
  pkt.set_length(n);

  n -= horde_net_packet::header_sz;
  bzero(pkt.get_data(), n);

  data_fragment_info* pheader = (data_fragment_info*)pkt.get_data();
  n -= sizeof (struct data_fragment_info);
  *pheader = data_fragment_info(0, 0, 0, n, n);

  consume_tx_slot(slot, pkt, 0, get_peer_address());
}

vector<data_fragment_info> 
scheduler_firehose::get_packet_contents(int packet_uid) {
  vector<data_fragment_info> empty;
  return empty;
}

//
// skeleton's functions
// 

#define MAX_HISTORY_SIZE 128
#define HISTORY_ERASE_SIZE 32

int
skeletal_scheduler_impl::packet_new(ptr<tx_slot> slot) {
  size_t n = slot->get_max_size();
  return packet_new(slot, n);
}

int
skeletal_scheduler_impl::packet_new(ptr<tx_slot> slot, size_t n) {
  if ((reuse_current) && (n <= current_size)) { 
    // we want to reuse this, and the slot is the right size
    current_used = horde_net_packet::header_sz;
  } else {
    // can't, or don't want to, reuse the current packet
    horde_net_packet new_pkt(n, PACKET_TYPE_DATA);
    
    current = new_pkt;

    current_size = n;
    current_used = horde_net_packet::header_sz;
    current_overhead = horde_net_packet::header_sz;
    
    current_uid = current_uid + 1;

    LOG_EVENT_DEBUG("debug") << "alloc new packet: " << (void*)new_pkt.get_buf().data() << ", uid:" << current_uid; 
  }

  vector<data_fragment_info> empty;
  current_contents = empty;

  reuse_current = false;

  return current_uid;
}

bool
skeletal_scheduler_impl::packet_is_full() {
  // packet is full when we can't add any more fragments...
  return (current_used > (current_size - sizeof(struct data_fragment_info) - 16));
}

bool 
skeletal_scheduler_impl::packet_has_data() {
  return current_used > horde_net_packet::header_sz;
}

constbuf
skeletal_scheduler_impl::packet_get_data() {
  return current.get_buf();
}

/// @todo better error handling
bool 
skeletal_scheduler_impl::packet_read_into(stream* stream) {
  char* pdata = current.get_buf().data() + current_used;
  
  // allocate a fragment header
  data_fragment_info* pheader = (data_fragment_info*)pdata;
  pdata += sizeof (struct data_fragment_info);
  
  // read metadata
  int adu_id, frag_offset, frag_len, adu_offset, adu_len;
  bool is_fragment;
  stream->sendq_read_metadata(&adu_id, &frag_offset, &frag_len, 
			      &is_fragment, &adu_offset, &adu_len, NULL); 
  int offset = frag_offset + adu_offset;

  // read in the data
  size_t maxread = current_size - (current_used + sizeof (struct data_fragment_info));
  int nread = stream->sendq_read_data(pdata, maxread);
  if (nread < 0) return true;  // error...

  //LOG_EVENT_DEBUG("debug") << "packet_read_into: read " << nread << " bytes: " << debugstr(pdata, nread);

  current_used += sizeof (struct data_fragment_info) + nread;
  current_overhead += sizeof (struct data_fragment_info);

  // set the header fields
  data_fragment_info inf(stream->get_remote_uid(), adu_id, offset, nread, adu_len);
  *pheader = inf; 
  inf.set_stream_id(stream->get_uid());
  current_contents.push_back(inf); // remember for later

  // log if debugging enabled
  LOG_EVENT_DEBUG("schedule") << "packet_read_into: stream=" << stream->get_uid()
			      << ", adu=" << adu_id << ", offs=" << offset << ", frag_offs=" << frag_offset 
			      << ", adu_len=" << adu_len << ", nread=" << nread;

  // return whether we're done with the packet...
  return (current_used > (current_size - sizeof(struct data_fragment_info) - 16));
}

bool 
skeletal_scheduler_impl::packet_read_into(int adu_id, dynbuf& buf, stream* stream) {
  char* pdata = current.get_buf().data() + current_used;  

  // allocate a fragment header
  data_fragment_info* pheader = (data_fragment_info*)pdata;
  pdata += sizeof (struct data_fragment_info);

  int adu_len = buf.length();

  // try to read in the data
  size_t maxread = current_size - (current_used + sizeof (struct data_fragment_info));
  if (maxread < adu_len) return false; // don't read in anything...

  memcpy(pdata, buf.data(), adu_len);
  current_used += sizeof (struct data_fragment_info) + adu_len;
  current_overhead += sizeof (struct data_fragment_info);

  // set the header fields
  data_fragment_info inf(stream->get_remote_uid(), adu_id, 0, adu_len, adu_len);
  *pheader = inf;
  inf.set_stream_id(stream->get_uid());
  current_contents.push_back(inf); // remember for later

  return (current_used > (current_size - sizeof(struct data_fragment_info) - 16));
}

bool 
skeletal_scheduler_impl::packet_write_int(int idata) { // big HACK
  char* pdata = current.get_buf().data() + current_used;  

  // allocate a fragment header
  data_fragment_info* pheader = (data_fragment_info*)pdata;
  pdata += sizeof (struct data_fragment_info);

  // set the header fields
  data_fragment_info inf(idata);
  *pheader = inf;

  return (current_used > (current_size - sizeof(struct data_fragment_info) - 16));
}

// assumes the packet has data
void 
skeletal_scheduler_impl::packet_send(ptr<tx_slot> slot) {
  // send the data
  current.set_length(current_used);
  consume_tx_slot(slot, current, current_uid, get_peer_address());

  //LOG_EVENT_DEBUG("debug") << "sent data: " << debugstr(current.get_data(), current.get_data_size());

  // remember contents in history
  send_history[current_uid] = current_contents;
  
  // prevent history from growing in an unbounded way.
  if (send_history.size() > MAX_HISTORY_SIZE) {
    map<int, vector<data_fragment_info> >::iterator low_erase = send_history.upper_bound(0);
    map<int, vector<data_fragment_info> >::iterator high_erase = send_history.lower_bound(low_erase->first+HISTORY_ERASE_SIZE);

    send_history.erase(low_erase, high_erase);
  }
}

vector<data_fragment_info> 
skeletal_scheduler_impl::get_packet_contents(int packet_uid) {
  map<int, vector<data_fragment_info> >::const_iterator iter = send_history.find(packet_uid);
  if (iter != send_history.end()) {
    return iter->second;
  } else {
    vector<data_fragment_info> empty;
    return empty;
  }
}

//
// round robin
//

void
scheduler_roundrobin::schedule_tx_slot(ptr<tx_slot> slot) {
  vector<stream*> streams = stream::get_connected_streams();
  int n = streams.size();

  if (n > 0) {
    int uid = packet_new(slot);
    LOG_EVENT_DEBUG("schedule") << "packing packet: " << uid << ", round:" << round;

    int looprevs = 0; // make sure there are no infinite loops...
    while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
      // one round (multiple rounds to pack a single packet are okay...)
      
      bool all_empty = true;
      
      for (int i = 0; i < n; i++) {
	int j = (i + round) % n;
	if (!streams[j]->sendq_empty()) {
	  all_empty = false;
	  if (packet_read_into(streams[j])) 
	    break; // packet is full
	}
      }
      
      if (all_empty) break; // break if no stream has data
    }

      LOG_EVENT("schedule") << "done with packet, packed " << packet_get_used() << " bytes, with " 
			    << packet_get_num_fragments() <<" fragments and " << packet_get_overhead() << "% overhead";
    
    if (packet_has_data()) packet_send(slot);
    else packet_reset();
    
    // give every stream a chance to be the first stream selected inside the for loop...
    round = (round + 1) % n;
  }
}

//
// hoard
//

void
scheduler_hoard::mtx_set_remove_worst() {
  map<int,mtx_set_elem_t>::iterator iter = mtx_set.upper_bound(0); // find lowest sequence number

  if (iter != mtx_set.end()) {
    mtx_set.erase(iter);
  }
}

void
scheduler_hoard::mtx_set_add(dynbuf& data, int adu, stream* pstream) {
  if (mtx_set.size() > MTX_SET_MAX_SIZE) {
    mtx_set_remove_worst();
  }

  mtx_set_seq++;
  mtx_set.insert(pair<int, mtx_set_elem_t>(mtx_set_seq, mtx_set_elem_t(data, adu, pstream)));
}

bool
scheduler_hoard::mtx_set_find_best_match(int length, dynbuf& buf, int& adu_id, stream* pstream) {
  // find the adu whose size is closest to length
  int best_seq = 0;
  int best_diff=length+1;
  for (map<int,mtx_set_elem_t>::iterator iter = mtx_set.begin(); iter != mtx_set.end(); iter++) {
    int diff = length - iter->second.data.length();
    if (diff >= 0) { // adu is smaller or equal in size
      if (diff < best_diff) best_seq = iter->first;
    }
  }

  if (best_seq) {
    buf = mtx_set[best_seq].data;
    adu_id = mtx_set[best_seq].adu_id;
    pstream = mtx_set[best_seq].pstream;
    mtx_set.erase(best_seq);
    return true;
  } else {
    return false;
  }
}

void
scheduler_hoard::schedule_tx_slot(ptr<tx_slot> slot) {
  vector<stream*> streams = stream::get_connected_streams();
  int n = streams.size();

  if (n > 0) {
    int uid = packet_new(slot);
    LOG_EVENT_DEBUG("schedule") << "packing packet: " << uid << ", round:" << round;

    set<int> groups_in_packet;

    int looprevs = 0; // make sure there are no infinite loops...
    while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
      // find the highest priority data we can pack into this packet,
      // given the constraints: loss threshold, correlation group,
      // fragmentation.
      stream* best_stream = NULL;
      horde_objective_tags best_tags;
      best_tags.priority = INT32_MAX;

      for (int i = 0; i < n; i++) {
	int j = (i + round) % n;
	stream* pstream = streams[j];
	if (!pstream->sendq_empty()) {
	  int totlen;
	  horde_objective_tags tags;	  
	  pstream->sendq_read_metadata(NULL, NULL, &totlen, NULL, NULL, NULL, &tags);
	  if (tags.priority < best_tags.priority) { // highest priority seen so far
	    if (tags.loss_threshold >= slot->get_loss_probability()) { // is the expected loss low enough...	      
	      if ((tags.correlation_group == 0) ||
		  (groups_in_packet.find(tags.correlation_group) == groups_in_packet.end())) {

		// correlation group matters, AND this data has a
		// different correlation group than other data in the
		// packet.
	     
		// can we fragment this data, or fit all of it into the current packet?
		size_t needed = sizeof (struct data_fragment_info) + totlen;
		if (((tags.fragmentation_allowed) && (totlen >= MIN_FRAGMENT_LEN))
		    || (packet_get_unused() >= needed)) {
		  best_stream = pstream;
		  best_tags = tags;
		} else { // group
		  LOG_EVENT_DEBUG("schedule") << "rejected stream " << i << ": fragmentation totlen=" << totlen << ", needed=" << needed;
		}
	      } else { // group
		LOG_EVENT_DEBUG("schedule") << "rejected stream " << i << ": correlation group " << tags.correlation_group;
	      }
	    } else { // loss
	      LOG_EVENT_DEBUG("schedule") << "rejected stream " << i << ": loss_threshold=" << tags.loss_threshold << ", loss="  << slot->get_loss_probability();
	    }
	  }
	}
      }

      groups_in_packet.insert(best_tags.correlation_group);
    
      if (best_stream) {
	// remember for multi tx?
	if (best_tags.multiple_tx && (!best_tags.fragmentation_allowed)) { 
	  int adu;
	  dynbuf buf = best_stream->sendq_front(&adu);
	  mtx_set_add(buf, adu, best_stream);
	}

	// read into packet
	if (packet_read_into(best_stream)) 
	  break; // packet is full      	
      } else {
	break; // nothing more can be packed into this slot.
      }    
    }

    // is there some space left for copies of the multiple_tx tagged packets?
    if (!mtx_set.empty()) {
      looprevs = 0; // make sure we don't loop infinitely.
      while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
	int remaining = packet_get_unused() - sizeof (struct data_fragment_info);

	int adu_id;
	dynbuf buf;
	stream* pstream; 
	if (mtx_set_find_best_match(remaining, buf, adu_id, pstream)) {
	  if (packet_read_into(adu_id, buf, pstream)) break;
	}
	else break; // no best match
      }
    }

    // if packet has data, send it...
    LOG_EVENT("schedule") << "done with packet, packed " << packet_get_used() << " bytes, with " 
			  << packet_get_num_fragments() <<" fragments and " << packet_get_overhead() << "% overhead";
    if (packet_has_data()) packet_send(slot);
    else packet_reset();

    // give every stream a chance to be the first stream selected inside the for loop...
    round = (round + 1) % n;
  }
}

//
// grunt
//
scheduler_grunt::scheduler_grunt() : skeletal_scheduler_impl(), round(0), groups_in_packet(), 
				     stream_srtt(), mtx_set_seq(0), mtx_set(), mtx_streams() {
  LOG_EVENT("schedule") << "created 'grunt' scheduler.";
  
  mtx_set_max_size = atoi(get_property("grunt_mtx_set_max_size", "-1").c_str());
  if (mtx_set_max_size < 0) mtx_set_max_size = MTX_SET_MAX_SIZE;

  LOG_EVENT_DEBUG("schedule") << "grunt::mtx_set_max_size=" << mtx_set_max_size;
}

void
scheduler_grunt::schedule_tx_slot(ptr<tx_slot> slot) {
  vector<stream*> streams = stream::get_connected_streams();
  int n = streams.size();

  if (n > 0) {
    // prepare for a new packet
    int uid = grunt_packet_new(slot);
    LOG_EVENT_DEBUG("schedule") << "packing packet: " << uid << ", round:" << round;

    // pack bits into this packet
    int looprevs = 0; // make sure there are no infinite loops...
    while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
      // find the highest priority data we can pack into this packet,
      // given the constraints.      
      
      stream *best_stream = NULL;
      horde_objective_tags best_tags;
      best_tags.priority = INT32_MAX;
      
      for (int i = 0; i < n; i++) {
	int j = (i + round) % n;
	stream *pstream = streams[i];
	LOG_EVENT_DEBUG("schedule") << j << ":'" << pstream->get_id() << "' size=" << pstream->get_sendq_size();

	if (!pstream->sendq_empty()) { // some data in stream's send queue
	  int totlen, offset; 
	  horde_objective_tags tags;	  
	  pstream->sendq_read_metadata(NULL, &offset, &totlen, NULL, NULL, NULL, &tags); // get adu info

	  if (tags.multiple_tx < 0) {
	    mtx_streams = true;
	    // maintain a certain size for this stream's queue
	    while (pstream->get_sendq_adu_count() > mtx_set_max_size) pstream->sendq_pop(); 

	  } else if (tags.priority < best_tags.priority) { // highest priority seen so far
	    if (grunt_packet_test_constraints(slot, pstream, totlen, offset, tags)) { // we can pack this data into this slot
	      best_stream = pstream;   // this is the best stream to read from
	      best_tags = tags;
	    }
	  }
	}
      }
      
      if (best_stream == NULL) break;    // if no stream has data that can go in this packet, we're done.
      else { 
	if (grunt_packet_read_into(slot, best_stream, best_tags)) // otherwise, read something into the packet.
	  break;                     // and exit when the packet is full
      }
    }

    // if unused space in packet, may have older bits to fill it up...
    grunt_packet_fill_unused_space();

    // if packet has data, send it
    LOG_EVENT("schedule") << "done with packet, packed " << packet_get_used() << " bytes, with " 
			  << packet_get_num_fragments() <<" fragments and " << packet_get_overhead() << "% overhead";
    
    if (packet_has_data()) packet_send(slot);
    else packet_reset();
    
    // give every stream a chance to be the first stream selected inside the for loop...
    //round = (round + 1) % n;
    round = rand() % n;
  }
}

int scheduler_grunt::grunt_packet_new(ptr<tx_slot> slot) {
  // alloc a new packet
  int uid = packet_new(slot);
  
  // reset state used by grunt_packet_test_constraints(...)
  groups_in_packet.clear();

  return uid;
}

bool scheduler_grunt::grunt_packet_read_into(ptr<tx_slot> slot, stream* src, horde_objective_tags& src_tags) {
  // remember for multi tx?
  if (src_tags.multiple_tx && (!src_tags.fragmentation_allowed)) {
    int adu;
    dynbuf buf = src->sendq_front(&adu);
    mtx_set_add(buf, adu, src);
  }
  
  // read data from stream queue
  bool full = packet_read_into(src);
  
  // remember which correlation groups are in this packet
  groups_in_packet.insert(src_tags.correlation_group);

  // update the stream's srtt
  int suid = src->get_uid();
  map<int,int>::iterator old_srtt = stream_srtt.find(suid);
  int ertt = slot->get_expected_rtt();
  if (old_srtt == stream_srtt.end()) {
    stream_srtt[suid] = ertt;
  } else {
    stream_srtt[suid] = (ertt + 7 * stream_srtt[suid]) / 8;
  }

  return full;
}

bool 
scheduler_grunt::grunt_packet_test_constraints(ptr<tx_slot> slot, stream* src, int data_len, int offset, horde_objective_tags& tags) {
  bool okay = true;

  // constraint I: size
  //     can we fragment this data, or fit all of it into the current packet?
  size_t needed = sizeof (struct data_fragment_info) + data_len - offset;
  size_t available = packet_get_unused();
  
  okay = (needed <= available);
  if (!okay) { // can't fit entire adu into the packet, can we fragment it?
    if (tags.fragmentation_allowed)
      okay = ((data_len >= MIN_FRAGMENT_LEN)
	      && ((available - sizeof (struct data_fragment_info)) >= MIN_FRAGMENT_LEN));
  }  
  if (!okay) return false;

  // constraint II: correlation group
  if (tags.correlation_group) {
    okay = (groups_in_packet.find(tags.correlation_group) == groups_in_packet.end());
  }
  if (!okay) return false;

  // constraint III: loss threshold
  okay = (tags.loss_threshold >= slot->get_loss_probability());
  if (!okay) return false;

  // constraint IV: latency threshold
  if (tags.rtt_threshold) {
    okay = slot->get_expected_rtt() < tags.rtt_threshold;
    if (!okay) return false;
  }

  // constraint V: relative loss threshold
  if (tags.relative_loss_threshold) {
    // calculate average loss over all channels
    int wloss_sum = 0, asrate_sum = 0;
    vector< ptr<channel_statistics> > stats = get_poolman()->get_statistics();
    for (vector< ptr<channel_statistics> >::const_iterator iter = stats.begin(); iter != stats.end(); iter++) {
      ptr<channel_statistics> stat = *iter;
      int loss = stat->get_loss_rate();
      int asrate = stat->get_allowed_send_rate(); 

      wloss_sum += loss * asrate;
      asrate_sum += asrate;
    }
    
    int ave_loss;
    if (asrate_sum > 0) ave_loss = wloss_sum / asrate_sum;
    else ave_loss = 0;

    int delta = slot->get_loss_probability() - ave_loss;
    if (delta > tags.relative_loss_threshold) okay = false;

    LOG_EVENT_DEBUG("grunt") << "test_constraints: ave_loss=" << ave_loss << ", delta=" << delta
			     << ", slotloss=" << slot->get_loss_probability() << ", thresh=" << tags.relative_loss_threshold;

    if (!okay) return false;
  }

  // constraint VI: relative latency threshold
  if (tags.relative_rtt_threshold) {
    int suid = src->get_uid();
    map<int,int>::iterator old_srtt = stream_srtt.find(suid);
    if (old_srtt != stream_srtt.end()) {
      int delta = slot->get_expected_rtt() - stream_srtt[suid];
      if (delta > tags.relative_rtt_threshold) okay = false;
    }
    if (!okay) return false;
  }

  // all constraints passed
  return true;
}

void 
scheduler_grunt::grunt_packet_fill_unused_space() {
  // is there some space left for copies of the (multiple_tx != 0) tagged packets?  
  // and are there some old multiple_tx>0 tagged packets or some multiple_tx<0 streams?
  if (!packet_is_full() && !mtx_set.empty()) {
    // handle multiple_tx < 0
    int looprevs = 0; // make sure we don't loop infinitely.
    while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
      int remaining = packet_get_unused() - sizeof (struct data_fragment_info);
      mtx_streams_read_into(remaining);
    }

    // handle multiple_tx > 0
    looprevs = 0;
    while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
      int remaining = packet_get_unused() - sizeof (struct data_fragment_info);
      
      int adu_id;
      dynbuf buf;
      stream* pstream; 
      if (mtx_set_find_best_match(remaining, buf, adu_id, pstream)) {
	LOG_EVENT_DEBUG("schedule") << "multiple_tx>0: adding adu " << adu_id << " from '" << pstream->get_id() << "'";
	if (packet_read_into(adu_id, buf, pstream)) break;
      }
      else break; // no best match
    }
  }
}

/// try to read one ADU from one of the MTX tagged streams.
bool
scheduler_grunt::mtx_streams_read_into(int maxsize) {
  if (mtx_streams) {
    // find a stream with an MTX<0 adu/frag that has size <= max_size
    vector<stream*> streams = stream::get_connected_streams();
    int n = streams.size();
    stream *selected=NULL;
    for (int i = 0; i < n; i++) {
      stream *pstream = streams[i];
      if (!pstream->sendq_empty()) { // some data in stream's send queue
	int totlen, offset; 
	horde_objective_tags tags;	  
	pstream->sendq_read_metadata(NULL, &offset, &totlen, NULL, NULL, NULL, &tags); // get adu info    
	
	if ((tags.multiple_tx < 0)) {
	  selected=pstream;
	  break;
	}
      }
    }
    
    if (selected == NULL) { // no MTX<0
      mtx_streams = false;
      return false;
    } else {
      LOG_EVENT_DEBUG("schedule") << "multiple_tx<0: reading from stream '" << selected->get_id() << "'";
      // pack bits from this stream into the packet
      packet_read_into(selected);
      return true;
    }
  }
}

void
scheduler_grunt::mtx_set_remove_worst() {
  map<int,mtx_set_elem_t>::iterator iter = mtx_set.upper_bound(0); // find lowest sequence number

  if (iter != mtx_set.end()) {
    mtx_set.erase(iter);
  }
}

void
scheduler_grunt::mtx_set_add(dynbuf& data, int adu, stream* pstream) {
  if (mtx_set.size() > mtx_set_max_size) {
    mtx_set_remove_worst();
  }

  mtx_set_seq++;
  mtx_set.insert(pair<int, mtx_set_elem_t>(mtx_set_seq, mtx_set_elem_t(data, adu, pstream)));
}

bool
scheduler_grunt::mtx_set_find_best_match(int length, dynbuf& buf, int& adu_id, stream*& pstream) {
  // find the adu whose size is closest to, and less than, length
  int best_seq = 0;
  int best_diff=length+1;
  for (map<int,mtx_set_elem_t>::iterator iter = mtx_set.begin(); iter != mtx_set.end(); iter++) {
    int diff = length - iter->second.data.length();
    if (diff >= 0) { // adu is smaller or equal in size
      if (diff < best_diff) best_seq = iter->first;
    }
  }

  if (best_seq) {
    buf = mtx_set[best_seq].data;
    adu_id = mtx_set[best_seq].adu_id;
    pstream = mtx_set[best_seq].pstream;

    mtx_set.erase(best_seq);
    return true;
  } else {
    return false;
  }
}

//
// coder
//
#define CODER_SCHED_BLOCK_SIZE (1024 - sizeof(int32_t))
#define CODER_SCHED_WINDOW_SIZE 3

scheduler_coder::scheduler_coder(): 
  round(0), coder(CODER_SCHED_BLOCK_SIZE, CODER_SCHED_WINDOW_SIZE), coded(NULL), skeletal_scheduler_impl() {
  LOG_EVENT("schedule") << "created 'coder' scheduler: block-size=" << CODER_SCHED_BLOCK_SIZE
			<< ", packet-sequence: (" << CODER_SCHED_WINDOW_SIZE << "-norm + 1-XOR)";
}

void
scheduler_coder::schedule_tx_slot(ptr<tx_slot> slot) {
  if (coded) _coded_send(slot);
  else {
    vector<stream*> streams = stream::get_connected_streams();
    int n = streams.size();
    
    if (n > 0) {
      int uid = packet_new(slot, CODER_SCHED_BLOCK_SIZE);
      packet_write_int(uid);
      LOG_EVENT_DEBUG("schedule") << "packing packet: " << uid << ", round:" << round;

      int looprevs = 0; // make sure there are no infinite loops...
      while (!packet_is_full() && (looprevs++ < SCHEDULE_TX_SLOT_MAX_LOOPREVS)) {
	// one round (multiple rounds to pack a single packet are okay...)
	
	bool all_empty = true;
	
	for (int i = 0; i < n; i++) {
	  int j = (i + round) % n;
	  if (!streams[j]->sendq_empty()) {
	    all_empty = false;
	    if (packet_read_into(streams[j])) 
	      break; // packet is full
	  }
	}
	
	if (all_empty) break; // break if no stream has data
      }
      
      LOG_EVENT("schedule") << "done with packet, packed " << packet_get_used() << " bytes, with " 
			    << packet_get_num_fragments() <<" fragments and " << packet_get_overhead() << "% overhead";
      
      if (packet_has_data()) {
	packet_send(slot);
	constbuf d = packet_get_data(); // including header
	pair<unsigned char*, size_t> p = coder.push((const unsigned char*)d.data(), current_uid, d.length());
	if (p.first) _coded_create(p);	 
      }

      // give every stream a chance to be the first stream selected inside the for loop...
      round = (round + 1) % n;
    }
  }
}

void
scheduler_coder::_coded_create(pair<unsigned char*, size_t> p) {
  coded = new horde_net_packet(CODER_SCHED_BLOCK_SIZE + horde_net_packet::header_sz + sizeof(int32_t), 
			       PACKET_TYPE_CODED_DATA); 
  current_uid = current_uid + 1;
  coded->set_length(p.second + horde_net_packet::header_sz);
  memcpy(coded->get_data()+sizeof(int32_t), p.first, p.second);
  int32_t* i = (int32_t*)(coded->get_data());
  *i = current_uid;
}

void
scheduler_coder::_coded_send(ptr<tx_slot> slot) {
  consume_tx_slot(slot, *coded, current_uid, get_peer_address());
  delete coded;
  coded = NULL;
}


HORDE_NAMESPACE_END;
