#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <execinfo.h>

#include <iostream>
#include <nmstl/ioevent>
#include <nmstl/net>

#include "properties.h"
#include "horde_log.h"

#include "session_daemon.h"
#include "syscall/syscall.h"
#include "session/session.h"
#include "net/poolman.h"
#include "core/core.h"

const char* DEFAULT_PROPERTIES_FILE = "horde.properties";
const char* DEFAULT_NAMESPACE_ROOT = "/var/horde";
const char* DEFAULT_INTERFACE_LIST = "eth0, eth1, ppp0, ppp1, ppp2, ppp3, tun0, tun1, tun2, tun3";
const char* DEFAULT_POOLMAN_SLEEP = "5000";
const char* DEFAULT_SCHEDULER = "default";
const char* DEFAULT_MANAGER_TYPE = "cbr{packet_size=1024, bitrate=40}";

io_event_loop io_loop;
string namespace_root;

const char *properties_filename = DEFAULT_PROPERTIES_FILE;

using namespace horde;

static void log_backtrace() {
  void *array[10];
  size_t size;
  char **strings;
  size_t i;

  size = backtrace (array, 10);
  strings = backtrace_symbols (array, size);

  LOG_EVENT("trace") << "STACK TRACE: " << size << " stack frames.";

  for (i = 0; i < size; i++)
    LOG_EVENT("trace") << i << ":\t" << strings[i];

  free (strings);
}

static void interrupt_signal(int s) {
  string sigstr;
  switch (s) {
  case SIGINT: sigstr = "SIGINT"; break;
  case SIGTERM: sigstr = "SIGTERM"; break;
  case SIGHUP: sigstr = "SIGHUP"; break;
  case SIGSEGV: sigstr = "SIGSEGV"; break;
  default: sigstr = "#" + s;
  }
  
  LOG_EVENT("signal") << "Received signal " << sigstr << ", exiting...";
  if (s == SIGSEGV) log_backtrace();     // log a backtrace  
  horde::exit();
}

// main function -- bootstraps daemon modules
main (int argc, char** argv) {
  //
  // handle command line arguments -- TODO: make this better...
  //
  bool accepting_session;
  string peer_address;
  int ippc_owner_fd;
  string session_id;
  
  if (argc != 6) {
    cerr << "Do not call this command directly: this is the horde session daemon, and must be initialized by the dispatch daemon.\n";
    exit(-1);  
  } else {
    properties_filename = argv[1];
    accepting_session = (strcmp(argv[2], "ACCEPT") == 0); 
    session_id = string(argv[3]);
    peer_address = string(argv[4]);
    ippc_owner_fd = atoi(argv[5]);
  }

  //
  // load properties file
  //
  if (!horde::load_properties_file(properties_filename)) {
    cerr << "Could not load properties file '" << properties_filename << "', using defaults..." << endl;
  }
  namespace_root = horde::get_property("namespace_root", DEFAULT_NAMESPACE_ROOT);

  //
  // initialize logging
  //
  if (!horde::logging_init(session_id.c_str(), namespace_root.c_str())) {
    // critical error
    cerr << "CRITICAL ERROR: could not create log files. Namespace root: " << namespace_root << endl
	 << "Exiting." << endl;
    exit(-1);
  }

  LOG_EVENT("properties") << horde::count_properties() << " unique properties loaded from file.";

  //
  // start IPPC server and set command
  //
  io_principal session_owner = 
    horde::init_syscall(io_loop, namespace_root + "/sessions/" + session_id + "/ippc", ippc_owner_fd);

  //
  // start network channel pool manager
  //
  int poolman_sleep_interval = atoi(get_property("poolman_sleep_interval", DEFAULT_POOLMAN_SLEEP).c_str());
  if (poolman_sleep_interval < 200) poolman_sleep_interval = 200;

  string iface_list = get_property("managed_interfaces", DEFAULT_INTERFACE_LIST);
  vector<string> ifaces = properties::decode_array_value(iface_list);

  vector< pair<string, string> > iface_params;

  for (vector<string>::const_iterator iter = ifaces.begin(); iter != ifaces.end(); iter++) {
    string type_str = get_property(*iter + "_manager_type", DEFAULT_MANAGER_TYPE);
    iface_params.push_back(pair<string, string>(*iter, type_str));
  }

  create_poolman(iface_params, poolman_sleep_interval, io_loop);
  network_pool_manager* poolman = get_poolman();

  if (poolman) {
    poolman->start();
  } else {   
    horde::panic("Could not create an interface pool manager.");
  }

  // 
  // initialize core
  //
  if (!horde::init_core(get_property("scheduler", DEFAULT_SCHEDULER), io_loop)) {
    CRIT << "couldn't create scheduler: "<< get_property("scheduler", DEFAULT_SCHEDULER);
    horde::panic("Could not create the requested scheduler");
  }

  //
  // initialize session
  // 
  horde::init_session(); // init session module
  session::initialize(session_id, peer_address, accepting_session, session_owner); // start session

  //
  // setup signal handlers
  //
  signal(SIGINT, interrupt_signal);
  signal(SIGTERM, interrupt_signal);
  signal(SIGHUP, interrupt_signal);
  signal(SIGSEGV, interrupt_signal);
  
  siginterrupt(SIGINT, 1);
  siginterrupt(SIGTERM, 1);
  siginterrupt(SIGHUP, 1);
  siginterrupt(SIGSEGV, 1);

  //
  // enter I/O loop
  //
  io_loop.run();

  LOG_EVENT("i/o") << "unexpected exit from I/O loop.";
  horde::exit();
}
