#include <stdlib.h>
#include <stdio.h>

#include "manager_impl.h"
#include "channel_sim.h"
//#include "manager_impl_xtra.h"

HORDE_NAMESPACE_BEGIN;

ptr<network_channel_manager>
create_network_channel_manager(string channel_name, int channel_id, 
		       string type_str, inet_address bind_address, io_event_loop& ioloop) {
  string type="", param_str="";
  string::size_type open_paren_idx = type_str.find("{");
  string::size_type close_paren_idx = type_str.find("}");
  
  if (open_paren_idx != string::npos) {
    type = to_lowercase(trim_spaces(type_str.substr(0, open_paren_idx)));
    param_str = type_str.substr(open_paren_idx+1, close_paren_idx - open_paren_idx - 1);
  } else {
    type = to_lowercase(trim_spaces(type_str));
  }
  
  properties params(param_str);

  network_channel_manager* man = NULL;

  if (type == "aimd") {
    man = new aimd_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
  } else if (type == "verb") {
    man = new verb_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
  } else if (type == "verb2") {
    man = new verb2_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
    //  } else if (type == "dccp") {
    //    man = new dccp_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
    //  } else if (type == "tcp") {
    //    man = new tcp_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
  } else {
    man = new cbr_channel_manager_impl(channel_name, channel_id, params, bind_address, ioloop);
  }

  return ptr<network_channel_manager>(man);
}

// 
// skeleton channel manager implementation
//
#define HEARTBEAT_PERIOD 500

skeleton_channel_manager_impl::skeleton_channel_manager_impl (string name, int id,
							      properties& params, 
							      inet_address bind_addr_proto, 
							      io_event_loop& ioloop)
  : channel_id(id), send_seq(0), 
    srtt_msec(-1), rttvar_msec(-1), rto_msec(5000), count_lost_bytes(0), last_count_lost(0), loss_rate(0.0),
    ack_count(0), loss_count(0), total_sent(0), timeouts(), map_packet_uid(), map_packet_size(),
    heartbeat_timer(NULL), ack_send_timer(NULL), ackq(), running(false), io_loop(&ioloop)
{
  //LOG_EVENT_DEBUG("debug") << params.to_string();

  int port = atoi(params.get("port", "-1").c_str());
  inet_address bind_address;
  if (port < 0) bind_address = inet_address(bind_addr_proto.get_host());
  else bind_address = inet_address(bind_addr_proto.get_host(), port);
  
  max_ack_delay = atoi(params.get("max_ack_delay", "0").c_str());
  if (max_ack_delay < 0) max_ack_delay = MAX_ACK_DELAY;

  if (params.contains("simulated")) { // is this managing a simulated channel?
    iohandler = ptr<channel_iohandler>(new simulated_channel_iohandler(params, ioloop, name, bind_address, 
								       wrap(this, &skeleton_channel_manager_impl::handle_recv),
								       wrap(this, &skeleton_channel_manager_impl::stat_update_beat)));
  } else {
    iohandler = ptr<channel_iohandler>(new channel_iohandler(ioloop, name, bind_address, 
							     wrap(this, &skeleton_channel_manager_impl::handle_recv),
							     wrap(this, &skeleton_channel_manager_impl::stat_update_beat)));
  }
}

bool
skeleton_channel_manager_impl::startup() {
  iohandler->start_io();
  schedule_heartbeat();
  running = true;
  return true;
}

bool 
skeleton_channel_manager_impl::shutdown() {
  iohandler->stop_io();
  stop_heartbeat();
  running = false;
  return true;
}

void 
skeleton_channel_manager_impl::schedule_heartbeat() {
  heartbeat_timer= (new callback_timer
		    (*(io_loop), wrap(this, &skeleton_channel_manager_impl::loss_detector_beat)));
  heartbeat_timer->arm(ntime::now_plus_msecs(HEARTBEAT_PERIOD));
}

void 
skeleton_channel_manager_impl::stop_heartbeat() {
 if (heartbeat_timer != NULL) { 
   heartbeat_timer->disarm();
   heartbeat_timer = NULL;
 }
}


uint32_t
skeleton_channel_manager_impl::update_rtt_estimators(uint32_t time_echo) {
  uint32_t now = time32_now();
  uint32_t rtt = now - time_echo;
  if (rtt <= 15000) { // 15 sec max... ignore higher (maybe <0 rtt)
    if (srtt_msec < 0) { // first rtt
      srtt_msec = rtt;
      rttvar_msec = rtt / 2;
      rto_msec = srtt_msec + max(HEARTBEAT_PERIOD, 4 * rttvar_msec);
    } else {
      rttvar_msec = (3 * rttvar_msec + abs((int)((int)srtt_msec - rtt))) / 4;
      srtt_msec = (7 * srtt_msec + rtt) / 8;
      rto_msec = srtt_msec + max(HEARTBEAT_PERIOD, 4 * rttvar_msec);
    }
    rto_msec *= 4; // @todo HACK
    if (rto_msec < 1000) rto_msec = 1000;
  }
  LOG_EVENT_DEBUG("net.rtt") << "update_rtt_estimators: srtt=" << srtt_msec 
			     << ", rttvar=" << rttvar_msec
			     << ", rto=" << rto_msec;
  return rtt;
}

void 
skeleton_channel_manager_impl::process_incoming_net_ack(packet_ack* ack) {
  int seq = ack->get_sequence();
  LOG_EVENT_DEBUG("net") << get_channel_name() << " processing ack: " << seq;
  ack_count++;
  map<int,int>::iterator iter = map_packet_uid.find(seq);
  if (iter != map_packet_uid.end()) { // if this is first event for this packet, continue
    // 1. update srtt
    uint32_t rtt = update_rtt_estimators(ack->timestamp_echo);
    
    // 2. use 3-DUPACK' rule to detect packet losses...
    //  rule: IF I see packet(i) then packet(i-3) is assumed lost
    //  why?: allows two later packets to arrive earlier than a(i)
    // impl: use the timeouts list
    int min_valid_seq = seq - 2 * MAX_ACKS_PER_PACKET;
    if ((!timeouts.empty()) && (timeouts.front().first < min_valid_seq)) {
      while((!timeouts.empty()) && (timeouts.front().first < min_valid_seq)) {
	int del_seq = timeouts.front().first;
	
	map<int,int>::iterator iter = map_packet_uid.find(del_seq);
	if (iter != map_packet_uid.end()) { // loss detected for an outstanding packet...
	  // handle loss...
	  loss_detected(del_seq, iter->second, false);
	
	  // delete from map_packet_uid
	  map_packet_uid.erase(iter);
	}
	
	timeouts.pop_front();
      }
    }
    
    // 3. deliver ack...
    int packet_uid = iter->second;
    notify_ack(seq, packet_uid, rtt);             // to subclass
    notify_core_ack_received(packet_uid, rtt);    // to source streams
    
    // 4. erase from maps
    map_packet_uid.erase(iter);                  // from seq->uid map
    map_packet_size.erase(packet_uid);            // from uid->size map

    // 5 .log
    //LOG_EVENT_DEBUG("net") << "ack maps to packet: " << packet_uid;    
    LOG_NET("ack") << get_channel_name() << "\t" << packet_uid << "\t" << rtt;

  } else {
    // else: ignore all packets for which ack/loss events have been generated
    // log?
    LOG_EVENT("net") << "WARNING: could not map an ack sequence to a sent packet: seq=" << seq;
  }
}

void
skeleton_channel_manager_impl::send_net_ack(horde_net_packet packet, inet_address dest_addr) {
  if (max_ack_delay > 0) {  // if some delay is allowed, put the ack on the ack queue
    packet_ack ack;
    packet_header* header = packet.get_header();
    
    ack.channel_id_seq = header->channel_id_seq;
    ack.timestamp_echo = header->timestamp;
    
    ackq.push_back(pair<inet_address, packet_ack>(dest_addr, ack));
    
    // figure out when to send the ack packet
    if (ack_send_timer == NULL) { // schedule a queue flush in the near future
      ack_send_timer = new callback_timer(*(io_loop), 
					  wrap(this, &skeleton_channel_manager_impl::send_composite_ack_now));
      
      ack_send_timer->arm(ntime::now_plus_msecs(max_ack_delay));
    } // otherwise, just wait for the timer to fire...
  } else { // send this ack immediately
    horde_net_packet ackp(horde_net_packet::header_sz + sizeof(struct packet_ack), PACKET_TYPE_COMPOSITE_ACK);
    packet_ack* ack = (packet_ack*)ackp.get_data();
    packet_header* header = packet.get_header();
    ack->channel_id_seq = header->channel_id_seq;
    ack->timestamp_echo = header->timestamp;

    send_control_packet(dest_addr, ackp);
  }
}


void 
skeleton_channel_manager_impl::process_incoming_composite_ack(horde_net_packet& packet) {
  packet_ack* ack = (packet_ack*)packet.get_data();
  void *end = packet.get_data() + packet.get_data_size();
  
  while (((void*)ack) < end) {
    process_incoming_net_ack(ack);
    ack++;
  }
}

void
skeleton_channel_manager_impl::send_composite_ack_now() {
  // produce one or more composite ack packets

  if (!ackq.empty()) {
    
    while (!ackq.empty()) {
      horde_net_packet ackp(horde_net_packet::header_sz + MAX_ACKS_PER_PACKET * sizeof(struct packet_ack), PACKET_TYPE_COMPOSITE_ACK);
      packet_ack* ack = (packet_ack*)ackp.get_data();
      int n = 0;

      list< pair<inet_address, packet_ack> >::iterator iter = ackq.begin();
      inet_address ack_dest_addr = iter->first;

      for (; iter != ackq.end(); iter++) {
	if (iter->first == ack_dest_addr) { // ack being sent to same peer...
	  *ack = iter->second;
	  ack++;
	  ackq.erase(iter);
	  n++;
	  if (n >= MAX_ACKS_PER_PACKET) break;
	}
      }

      ackp.set_length(horde_net_packet::header_sz + n * sizeof(struct packet_ack));

      LOG_EVENT_DEBUG("net") << "sending composite ack x" << n;
      send_control_packet(ack_dest_addr, ackp);
    }
  }

  ack_send_timer = NULL;
}

void
skeleton_channel_manager_impl::handle_recv(horde_net_packet recv_data, inet_address recv_addr) {
  if (recv_data.get_type() == PACKET_TYPE_COMPOSITE_ACK) {
    process_incoming_composite_ack(recv_data);     // process localy
  } else {
    notify_core_packet_received(recv_data, recv_addr, channel_id);   // pass on to core

    if (recv_data.get_type() == PACKET_TYPE_DATA) {
      send_net_ack(recv_data, recv_addr);
    }
  }
}

void
skeleton_channel_manager_impl::send_data(horde_net_packet data, int packet_uid, inet_address send_addr) {
  // tag packet header
  packet_header* hdr = data.get_header();
  hdr->set_channel_id(channel_id);
  hdr->set_sequence(send_seq++);
  uint32_t now = time32_now();
  hdr->timestamp = now;

  // debug: log packet
  LOG_EVENT_DEBUG("net") << get_channel_name() << ": consuming slot, channel_seq=" << hdr->get_sequence()
			 << ", packet UID:" << packet_uid <<", len:" << data.get_length(); 

  // store packet_info in seq->info map..
  int seq = hdr->get_sequence();
  map_packet_uid[seq] = packet_uid;
  map_packet_size[packet_uid] = data.get_length();

  // set up a timeout for this packet...
  timeouts.push_back(pair<int,int>(seq, now + rto_msec));

  // update count
  total_sent++;

  // actually send the packet...
  iohandler->async_send(data, send_addr);
}

void
skeleton_channel_manager_impl::loss_detector_beat() {
  if (!timeouts.empty()) {
    // All packets with expired timers in map_packet_uid are flagged
    // as losses...
    uint32_t now = time32_now();
        
    while((!timeouts.empty()) && (timeouts.front().second <= now)) {
      int del_seq = timeouts.front().first;

      map<int,int>::iterator iter = map_packet_uid.find(del_seq);
      if (iter != map_packet_uid.end()) { // loss detected for an outstanding packet...
	// handle loss...
	loss_detected(del_seq, iter->second, true);

	// delete from map_packet_uid
	map_packet_uid.erase(iter);
      }

      timeouts.pop_front();
    }
  }

  if (running) schedule_heartbeat();
}

void
skeleton_channel_manager_impl::loss_detected(int seq, int packet_uid, bool timeout_loss) {
  // log
  LOG_NET("loss") << get_channel_name() << "\t" << packet_uid << "\t" 
		  << (timeout_loss ? "TIMEOUT" : "missing ACK");

  // update statistics
  loss_count++;
  map<int,int>::iterator i = map_packet_size.find(packet_uid);
  if (i != map_packet_size.end()) {
    int sz = i->second;
    count_lost_bytes += sz;
  }

  // notify other modules of this loss
  notify_loss(seq, packet_uid, timeout_loss);   // first notify subclass
  notify_core_loss_detected(packet_uid);        // then notify the core
}

void 
skeleton_channel_manager_impl::send_control_packet(inet_address send_addr, horde_net_packet data, bool no_delay) {
  if (no_delay) 
    iohandler->send_now(data, send_addr);
  else
    iohandler->async_send(data, send_addr);
}

void 
skeleton_channel_manager_impl::stat_update_beat() {
  // update loss statistics
  int sent = iohandler->get_ave_send_rate();
  int lost = count_lost_bytes - last_count_lost;

  float lrate = 0.0;
  if (sent > 0) lrate = float(lost) / sent;
  lrate *= 100;

  loss_rate = (3 * loss_rate + lrate) / 4;

  last_count_lost = count_lost_bytes;

  LOG_EVENT_DEBUG("net") << "stat_update_beat: count_lost_bytes=" << count_lost_bytes << ", lost=" << lost << ", sent=" << sent << ", lrate=" << lrate << ", loss_rate=" << loss_rate;

  // child may need to update statistics
  this->statistics_update();
}

//
// CBR manager
// 

cbr_channel_manager_impl::cbr_channel_manager_impl (string name, int id, properties& params, 
						    inet_address bind_addr_proto, 
						    io_event_loop& ioloop) 
  : slot_gen_timer(NULL),
    skeleton_channel_manager_impl(name, id, params, bind_addr_proto, ioloop)
{
  packet_size = atoi(params.get("packet_size", "1024").c_str());
  if (packet_size <= 0) packet_size = 1024;

  bitrate = atoi(params.get("bitrate", "20").c_str());
  if (bitrate <= 0) bitrate = 1;

  period_msecs = (packet_size * 1000) / ((bitrate * 1024) / 8);
  if (period_msecs > 10) period_msecs = 10;

  LOG_EVENT("net") << get_channel_name() << ": created cbr channel manager: bitrate=" << bitrate
		   << " kbps, packet_size=" << packet_size
		   << ", address=" << get_address()
		   << ", period_msecs=" << period_msecs;

  //LOG_EVENT_DEBUG("debug") << params.to_string();
}



bool
cbr_channel_manager_impl::startup() {
  bool x = skeleton_channel_manager_impl::startup();

  if (x) slot_gen_timer_fire();

  return x;
}

bool
cbr_channel_manager_impl::shutdown() { 
  bool x = skeleton_channel_manager_impl::shutdown();

  if (slot_gen_timer != NULL) {
    slot_gen_timer->disarm();
    slot_gen_timer = NULL;
  }

  return x;
}

void
cbr_channel_manager_impl::consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid,  inet_address send_address) {
  if (slot->is_valid()) {
    // send the data    
    send_data(data, packet_uid, send_address);
  } else {
    // notify core about the loss    
    notify_core_loss_detected(packet_uid);
  }
}

 
void 
cbr_channel_manager_impl::slot_gen_timer_fire () {
  // generate a slot
  LOG_EVENT_DEBUG("net") << get_channel_name() << ": generating tx slot.";
  
  ptr<tx_slot> slot = ptr<tx_slot>(new tx_slot(this, get_average_rtt_msec(), get_loss_rate(), packet_size));
  notify_core_new_tx_slot(slot);
  
  // reschedule self...   
  slot_gen_timer = (new callback_timer
		    (*(io_loop), wrap(this, &cbr_channel_manager_impl::slot_gen_timer_fire)));

  slot_gen_timer->arm(ntime::now_plus_msecs(period_msecs));
}

int
cbr_channel_manager_impl::get_allowed_send_rate() {
  return ( bitrate * 1024 ) / 8;
}

//
// AIMD channel manager
//
aimd_channel_manager_impl::aimd_channel_manager_impl (string name, int id, properties& params, 
						      inet_address bind_addr_proto, 
						      io_event_loop& ioloop) 
  : cwnd(0), fract_cwnd(0), last_loss_seq(-123), state(SLOW_START),
    slot_gen_timer(NULL),
    skeleton_channel_manager_impl(name, id, params, bind_addr_proto, ioloop)
{
  packet_size = atoi(params.get("packet_size", "1024").c_str());
  if (packet_size <= 0) packet_size = 1024;

  max_cwnd = atoi(params.get("max_cwnd", "-1").c_str());
  if (max_cwnd < 0) max_cwnd = DEFAULT_MAX_CWND;

  LOG_EVENT("net") << "created aimd channel manager: packet_size=" << packet_size
		   << ", max_cwnd=" << max_cwnd
		   << ", address=" << get_address();
}


 void
 aimd_channel_manager_impl::consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address) {
  if (slot->is_valid()) {
    // send the data    
    send_data(data, packet_uid, send_address);
  } else {
    notify_core_loss_detected(packet_uid);
  }
}

bool 
aimd_channel_manager_impl::startup() {
  bool x = skeleton_channel_manager_impl::startup();

  if (x) {
    cwnd = 1;
    fract_cwnd = 0;
    state = SLOW_START ;

    //generate_tx_slot();
    slot_gen_timer_fire();
  }

  return x;
}

bool
aimd_channel_manager_impl::shutdown() { 
  bool x = skeleton_channel_manager_impl::shutdown();

  if (slot_gen_timer != NULL) {
    slot_gen_timer->disarm();
    slot_gen_timer = NULL;
  }

  return x;
}

void
aimd_channel_manager_impl::generate_tx_slot() {
  LOG_EVENT_DEBUG("net") << get_channel_name() << ": generating tx slot.";

  ptr<tx_slot> slot = ptr<tx_slot>(new tx_slot(this, get_average_rtt_msec(), get_loss_rate(), packet_size));				   
  notify_core_new_tx_slot(slot);
}

void 
aimd_channel_manager_impl::slot_gen_timer_fire () {
  // generate a slot
  generate_tx_slot();
  
  // reschedule self...   
  slot_gen_timer = (new callback_timer
		    (*(io_loop), wrap(this, &aimd_channel_manager_impl::slot_gen_timer_fire)));

  slot_gen_timer->arm(ntime::now_plus_msecs(SLOT_GEN_TIMER_PERIOD));
}

/// @todo this is kind of broken...
void 
aimd_channel_manager_impl::notify_loss (int channel_seq, int packet_uid, bool timeout_loss) {

  if (channel_seq = last_loss_seq+1) { // treat burst losses as one loss
    // halve window...
    state = AIMD;
    fract_cwnd += cwnd * 1000;
    fract_cwnd /= 2;
    cwnd = fract_cwnd / 1000;
    fract_cwnd = fract_cwnd % 1000;

    // one slot for the whole burst
    if (cwnd < 1) { 
      cwnd = 1;
      fract_cwnd = 0;
    }

    // the rest will be generated by acks, but generate one slot
    // just in case everything in-transit was lost...
    generate_tx_slot();
  }

  last_loss_seq = channel_seq;
}

void 
aimd_channel_manager_impl::notify_ack (int channel_seq, int packet_uid, int rtt_msec) {
  // regenerate tx slot
  generate_tx_slot();

  // generate an additional slot?
  if (cwnd < max_cwnd) {
    if (state == AIMD) {
      // grow cwnd
      fract_cwnd += 1000/get_average_rtt_msec();
      
      if (fract_cwnd > 1000) {
	fract_cwnd -= 1000;
	cwnd++;
	
	generate_tx_slot();
      }
    } else if (state == SLOW_START) {
      cwnd++;
      
      generate_tx_slot();
    }
  }
}

int
aimd_channel_manager_impl::get_allowed_send_rate() {
  int srtt = get_average_rtt_msec();
  if (srtt < 1) srtt = 1;

  int rate = (cwnd * packet_size) / srtt;
  if (rate <= packet_size) rate = packet_size;
  
  return rate;
}

//
// verb implementation
//
verb_channel_manager_impl::verb_channel_manager_impl (string name, int id, properties& params, 
						      inet_address bind_addr_proto, 
						      io_event_loop& ioloop) 
  : cwnd(0.0), last_loss_seq(-123),in_flight(0), slow_start(true),
    slot_gen_timer(NULL), long_timer(NULL), short_timer(NULL), allowed_rate_ewma(0),
    skeleton_channel_manager_impl(name, id, params, bind_addr_proto, ioloop)
{
  packet_size = atoi(params.get("packet_size", "1024").c_str());
  if (packet_size <= 0) packet_size = 1024;

  long_timer_period = atoi(params.get("long_timer_period", "-1").c_str());
  if (long_timer_period < 0) long_timer_period = LONG_TIMER_PERIOD;

  maximum_min_rtt = atoi(params.get("maximum_min_rtt", "-1").c_str());
  if (maximum_min_rtt < 0) maximum_min_rtt = MAXIMUM_MIN_RTT;
  min_rtt = maximum_min_rtt + 1;

  min_probe_factor = float(atof(params.get("min_probe_factor", "-1").c_str()));
  if (min_probe_factor < 0) min_probe_factor = MIN_PROBE_FACTOR;

  max_probe_factor = float(atof(params.get("max_probe_factor", "-1").c_str()));
  if (max_probe_factor < 0) max_probe_factor = MAX_PROBE_FACTOR;
  if (max_probe_factor < min_probe_factor) max_probe_factor = min_probe_factor;
  probe_factor = max_probe_factor;

  probe_factor_step = float(atof(params.get("probe_factor_step", "-1").c_str()));
  if (probe_factor_step < 0) probe_factor_step = PROBE_FACTOR_STEP;

  max_send_per_second = atoi(params.get("max_send_per_second", "-1").c_str());
  if (max_send_per_second < 0) max_send_per_second = MAX_SEND_PER_SECOND;

  max_cwnd = float(atof(params.get("max_cwnd", "-1").c_str()));
  if (max_cwnd < 0) max_cwnd = DEFAULT_MAX_CWND;

  LOG_EVENT("net") << "created verb channel manager: address=" << get_address();
  LOG_EVENT_DEBUG("net") << "probe: max=" << float_to_str(max_probe_factor) 
			 << ", min=" << float_to_str(min_probe_factor) 
			 << ", step=" << float_to_str(probe_factor_step);
}




void
verb_channel_manager_impl::consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address) {
  if (slot->is_valid()) {
    // send the data    
    send_data(data, packet_uid, send_address);
    in_flight++;  // in transit...
  } else {
    notify_core_loss_detected(packet_uid);
  }
}

bool 
verb_channel_manager_impl::startup() {
  bool x = skeleton_channel_manager_impl::startup();

  if (x) {
    cwnd = 1.0;
    probe_factor = max_probe_factor;
    in_flight = 0;
    min_rtt = INITIAL_MIN_RTT;

    slot_gen_timer_fire();

    schedule_short_timer();
    schedule_long_timer(); 
  }

  return x;
}

bool
verb_channel_manager_impl::shutdown() { 
  bool x = skeleton_channel_manager_impl::shutdown();

  if (slot_gen_timer != NULL) {
    slot_gen_timer->disarm();
    slot_gen_timer = NULL;
  }

  if (short_timer != NULL) {
    short_timer->disarm();
    short_timer = NULL;
  }

  if (long_timer != NULL) {
    long_timer->disarm();
    long_timer = NULL;
  }

  return x;
}

void
verb_channel_manager_impl::generate_tx_slot() {
  if ((max_send_per_second > 0) && (max_send_per_second > iohandler->get_sent_count_current_second())) {
    LOG_EVENT_DEBUG("verb") << get_channel_name() << ": generating tx slot, sent:" << iohandler->get_sent_count_current_second();;

    ptr<tx_slot> slot = ptr<tx_slot>(new tx_slot(this, get_average_rtt_msec(), get_loss_rate(), packet_size));
    notify_core_new_tx_slot(slot);
    
    last_gen_time = ntime::now();
  } else {
    LOG_EVENT_DEBUG("verb") << "generate_tx_slot: too many bytes sent this second, not generating tx slot, sent:"
			    << iohandler->get_sent_count_current_second();
  }
}

void 
verb_channel_manager_impl::long_timer_fire () {
  if ((ntime::now() - last_gen_time).to_msecs() >= LONG_TIMER_PERIOD) {
    cwnd = 1.0;
    in_flight = 0;
    
    generate_tx_slot();
  }

  schedule_long_timer();
}

void 
verb_channel_manager_impl::schedule_long_timer () {
  // reschedule self...   
  long_timer = (new callback_timer
		 (*(io_loop), wrap(this, &verb_channel_manager_impl::long_timer_fire)));
  
  long_timer->arm(ntime::now_plus_msecs(long_timer_period));
}

void 
verb_channel_manager_impl::short_timer_fire () {
  if (min_rtt > maximum_min_rtt) {
    cwnd = (6 * cwnd) / 8;
    if (cwnd < 1.0) cwnd = 1.0;

    probe_factor -= probe_factor_step;
    if (probe_factor < min_probe_factor) probe_factor = min_probe_factor;      
  } else {
    probe_factor += probe_factor_step;
    if (probe_factor > max_probe_factor) probe_factor = max_probe_factor;
  }

  LOG_EVENT_DEBUG("verb") << "probe_factor=" << probe_factor << ", " 
			  << float_to_str(probe_factor) << ", " << float_to_str(cwnd);

  LOG_EVENT_DEBUG("verb") << "short_timer_fire: min_rtt=" << min_rtt
			  << ", cwnd=" << float_to_str(cwnd) 
			  << ", probe_factor=" << float_to_str(probe_factor)
			  << ", in_flight=" << in_flight;

  min_rtt = INITIAL_MIN_RTT;

  schedule_short_timer();
}

void 
verb_channel_manager_impl::schedule_short_timer () {
  int timeout = 2 * min_rtt;
  if (timeout > LONG_TIMER_PERIOD) timeout = LONG_TIMER_PERIOD;

  // reschedule self...   
  short_timer = (new callback_timer
		 (*(io_loop), wrap(this, &verb_channel_manager_impl::short_timer_fire)));
  
  short_timer->arm(ntime::now_plus_msecs(timeout));
}


void 
verb_channel_manager_impl::statistics_update() {
  // update allowed rate average
  allowed_rate_ewma += get_average_send_rate();
  allowed_rate_ewma /= 2;
  if (allowed_rate_ewma < packet_size)  allowed_rate_ewma = packet_size;
}

void 
verb_channel_manager_impl::slot_gen_timer_fire () {
  // generate a tx slot?
  if (cwnd > in_flight) generate_tx_slot();
  
  // reschedule self...   
  slot_gen_timer = (new callback_timer
		    (*(io_loop), wrap(this, &verb_channel_manager_impl::slot_gen_timer_fire)));
  
  slot_gen_timer->arm(ntime::now_plus_msecs(SLOT_GEN_TIMER_PERIOD));
}

void 
verb_channel_manager_impl::notify_loss (int channel_seq, int packet_uid, bool timeout_loss) {
  // exit slow start
  slow_start = false;

  // update cwnd
  if (channel_seq = last_loss_seq+1) { // treat burst losses as one loss
    // halve window...
    cwnd /= 2;
    if (cwnd < 1.0) cwnd = 1.0;
    
    in_flight = 0; // forget everything in flight...
  }

  // debug logging
  LOG_EVENT_DEBUG("verb") << "notify_loss: min_rtt=" << min_rtt
			  << ", cwnd=" << float_to_str(cwnd) 
			  << ", probe_factor=" << float_to_str(probe_factor)
			  << ", in_flight=" << in_flight;

  // remember this sequence
  last_loss_seq = channel_seq;
}

void 
verb_channel_manager_impl::notify_ack (int channel_seq, int packet_uid, int rtt_msec) {
  // keep track of minimum RTT
  if (rtt_msec < min_rtt) min_rtt = rtt_msec;

  // got one packet through, so in_flight decremented
  in_flight--;
  if (in_flight < 0) in_flight = 0;

  // grow the congestion window
  if (cwnd < max_cwnd) {
    if (slow_start) {
      cwnd += 1;
    } else {
      float div = cwnd;
      if (div < 1.0) div = 1.0;
      cwnd += probe_factor / div;
    }
  }

  // debug logging
  LOG_EVENT_DEBUG("verb") << "notify_ack: min_rtt=" << min_rtt
			  << ", cwnd=" << float_to_str(cwnd) 
			  << ", probe_factor=" << float_to_str(probe_factor)
			  << ", in_flight=" << in_flight;

  // generate some packets...
  int n = (int)(cwnd - in_flight);
  if (n > MAX_SLOT_GEN) n = MAX_SLOT_GEN;

  if ((n < 2) && (n >= 1))  {
    generate_tx_slot(); // most frequent case?
  } else {
    for (int i = 0; i < n; i++) generate_tx_slot();
  }
}

int
verb_channel_manager_impl::get_allowed_send_rate() {
  return allowed_rate_ewma;
}

//
// verb2 manager
// 
verb2_channel_manager_impl::verb2_channel_manager_impl (string name, int id, properties& params, 
						      inet_address bind_addr_proto, 
						      io_event_loop& ioloop) 
  : cwnd(0.0), last_loss_seq(-123),in_flight(0), slow_start(true),
    slot_gen_timer(NULL), long_timer(NULL), allowed_rate_ewma(0),
    skeleton_channel_manager_impl(name, id, params, bind_addr_proto, ioloop)
{
  packet_size = atoi(params.get("packet_size", "1024").c_str());
  if (packet_size <= 0) packet_size = 1024;

  long_timer_period = atoi(params.get("long_timer_period", "-1").c_str());
  if (long_timer_period < 0) long_timer_period = LONG_TIMER_PERIOD;

  probe_factor = float(atof(params.get("probe_factor", "-1").c_str()));
  if (probe_factor < 0) probe_factor = PROBE_FACTOR;

  max_send_per_second = atoi(params.get("max_send_per_second", "-1").c_str());
  if (max_send_per_second < 0) max_send_per_second = MAX_SEND_PER_SECOND;

  max_cwnd = float(atof(params.get("max_cwnd", "-1").c_str()));
  if (max_cwnd < 0) max_cwnd = DEFAULT_MAX_CWND;

  LOG_EVENT("net") << "created verb2 channel manager: address=" << get_address();
  LOG_EVENT_DEBUG("net") << "max_cwnd=" << max_cwnd << ", probe_factor=" << float_to_str(probe_factor);
}




void
verb2_channel_manager_impl::consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address) {
  if (slot->is_valid()) {
    // send the data    
    send_data(data, packet_uid, send_address);
    in_flight++;  // in transit...
  } else {
    notify_core_loss_detected(packet_uid);
  }
}

bool 
verb2_channel_manager_impl::startup() {
  bool x = skeleton_channel_manager_impl::startup();

  if (x) {
    cwnd = 1.0;
    in_flight = 0;

    slot_gen_timer_fire();

    schedule_long_timer(); 
  }

  return x;
}

bool
verb2_channel_manager_impl::shutdown() { 
  bool x = skeleton_channel_manager_impl::shutdown();

  if (slot_gen_timer != NULL) {
    slot_gen_timer->disarm();
    slot_gen_timer = NULL;
  }

  if (long_timer != NULL) {
    long_timer->disarm();
    long_timer = NULL;
  }

  return x;
}

void
verb2_channel_manager_impl::generate_tx_slot() {
  if ((max_send_per_second > 0) && (max_send_per_second > iohandler->get_sent_count_current_second())) {
    LOG_EVENT_DEBUG("verb2") << get_channel_name() << ": generating tx slot, sent:" << iohandler->get_sent_count_current_second();;

    ptr<tx_slot> slot = ptr<tx_slot>(new tx_slot(this, get_average_rtt_msec(), get_loss_rate(), packet_size));
    notify_core_new_tx_slot(slot);
    
    last_gen_time = ntime::now();
  } else {
    LOG_EVENT_DEBUG("verb2") << "generate_tx_slot: too many bytes sent this second, not generating tx slot, sent:"
			    << iohandler->get_sent_count_current_second();
  }
}

void 
verb2_channel_manager_impl::long_timer_fire () {
  if ((ntime::now() - last_gen_time).to_msecs() >= LONG_TIMER_PERIOD) {
    cwnd = 1.0;
    in_flight = 0;
    
    generate_tx_slot();
  }

  schedule_long_timer();
}

void 
verb2_channel_manager_impl::schedule_long_timer () {
  // reschedule self...   
  long_timer = (new callback_timer
		 (*(io_loop), wrap(this, &verb2_channel_manager_impl::long_timer_fire)));
  
  long_timer->arm(ntime::now_plus_msecs(long_timer_period));
}

void 
verb2_channel_manager_impl::statistics_update() {
  // update allowed rate average
  allowed_rate_ewma += get_average_send_rate();
  allowed_rate_ewma /= 2;
  if (allowed_rate_ewma < packet_size)  allowed_rate_ewma = packet_size;
}

void 
verb2_channel_manager_impl::slot_gen_timer_fire () {
  // generate a tx slot?
  if (cwnd > in_flight) generate_tx_slot();
  
  // reschedule self...   
  slot_gen_timer = (new callback_timer
		    (*(io_loop), wrap(this, &verb2_channel_manager_impl::slot_gen_timer_fire)));
  
  slot_gen_timer->arm(ntime::now_plus_msecs(SLOT_GEN_TIMER_PERIOD));
}

void 
verb2_channel_manager_impl::notify_loss (int channel_seq, int packet_uid, bool timeout_loss) {
  // exit slow start
  slow_start = false;

  // update cwnd
  if (channel_seq = last_loss_seq+1) { // treat burst losses as one loss
    // halve window...
    cwnd /= 2;
    if (cwnd < 1.0) cwnd = 1.0;
    
    in_flight = 0; // forget everything in flight...
  }

  // debug logging
  LOG_EVENT_DEBUG("verb2") << "notify_loss: cwnd=" << float_to_str(cwnd) 
			  << ", probe_factor=" << float_to_str(probe_factor)
			  << ", in_flight=" << in_flight;

  // remember this sequence
  last_loss_seq = channel_seq;
}

void 
verb2_channel_manager_impl::notify_ack (int channel_seq, int packet_uid, int rtt_msec) {
  // got one packet through, so in_flight decremented
  in_flight--;
  if (in_flight < 0) in_flight = 0;

  // grow the congestion window
  if (cwnd < max_cwnd) {
    if (slow_start) {
      cwnd += 1;
    } else {
      float div = cwnd;
      if (div < 1.0) div = 1.0;
      cwnd += probe_factor / div;
    }
  }

  // debug logging
  LOG_EVENT_DEBUG("verb2") << "notify_ack: cwnd=" << float_to_str(cwnd) 
			  << ", probe_factor=" << float_to_str(probe_factor)
			  << ", in_flight=" << in_flight;

  // generate some packets...
  int n = (int)(cwnd - in_flight);
  if (n > MAX_SLOT_GEN) n = MAX_SLOT_GEN;

  if ((n < 2) && (n >= 1))  {
    generate_tx_slot(); // most frequent case?
  } else {
    for (int i = 0; i < n; i++) generate_tx_slot();
  }
}

int
verb2_channel_manager_impl::get_allowed_send_rate() {
  return allowed_rate_ewma;
}


HORDE_NAMESPACE_END;
