#include "session.h"

#include "event_handlers.h"
#include "core/control_netio.h"
#include "horde_log.h"

#include <nmstl/net>
#include <nmstl/ioevent>

using namespace horde::session;
using namespace nmstl;

extern io_event_loop io_loop;

HORDE_NAMESPACE_BEGIN;

extern int xmux_recv_count; /// @todo this is a bit of a hack...

// module initialization

void init_session() {
  // initialize random number generator (for stream uid generation)
  srand( time(NULL) );
}

namespace session {
  // 
  // session state
  // 

  bool acceptor_session = false;
  bool connected = false;
  string session_id = "";
  io_principal session_owner = NULL;

  io_principal get_owner() {
    return session_owner;
  }

  string get_id() {
    return session_id;
  }

  bool is_connected() {
    return connected;
  }

  const int REPEAT_PERIOD_MSECS = 1500;

  void repeated_send_connect_request(inet_address peer_address) {
    if (!connected) {
      // send a connect request
      control_netio::send_session_connect_request(session_id, peer_address);

      // try again after a second
      (new callback_timer(io_loop, wrap(repeated_send_connect_request, peer_address)))->arm(ntime::now_plus_msecs(REPEAT_PERIOD_MSECS));
    }
  }

  void repeated_send_connect_ack(inet_address peer_address) {
    if (xmux_recv_count == 0) { // if we haven't received a request, assume last ack didn't get to peer.
      // send a connect request
      control_netio::send_session_connect_ack(session_id, peer_address);

      // try again after a second
      (new callback_timer(io_loop, wrap(repeated_send_connect_ack, peer_address)))->arm(ntime::now_plus_msecs(REPEAT_PERIOD_MSECS));
    }
  }

  void 
  initialize(string session_name, string peer_address, bool acceptor, io_principal owner) {
    if (!connected) {
      LOG_EVENT("session") << "init session: peer=" << peer_address << ", accept=" << acceptor;
    
      acceptor_session = acceptor;
      session_id = session_name;
      session_owner = owner;

      if (acceptor) {
	control_netio::set_session_info(session_id, peer_address);
	connected = true;

	repeated_send_connect_ack(inet_address(peer_address));

	notify_syscall_session_connected(owner);
	notify_core_session_connected();

      } else {
	repeated_send_connect_request(inet_address(peer_address));
      }

    } else {
      // should never get here: initialize() is only called during bootstrapping...
    }
  }

  void 
  join(string session_id, io_principal principal) {
    if (connected) {
      // okay, join succeeded
      notify_syscall_session_connected(principal);
    } else {
      // exception
      notify_syscall_session_exception(principal, EX_SESSION_ILLEGAL_OPERATION, "cannot join a session that isn't connected.");
    }
  }

  void 
  close(string session_id, io_principal principal) {
    if ((principal == session_owner) && (connected)) {
      LOG_EVENT("session") << "locally requested session close. closing session.";

      control_netio::send_session_close();
      connected = false;

      notify_syscall_session_closed(session_owner);   
    } else {
      // exception
      notify_syscall_session_exception(principal, EX_SESSION_INSUFFICIENT_PERMISSION, "only owner process can close the session.");
    }
  }

};

//
// network event handlers
//
void
notify_session_close_request(string req_session_id) {
  if ((session_id == req_session_id) && (connected)) {
    LOG_EVENT("session") << "network peer sent close message. closing session.";

    connected = false;
    notify_syscall_session_closed(session_owner, true);
  }
}


void 
notify_session_connect_request(string session_id, inet_address peer) {  
  // last ack must have dropped, so reply with an ack if the session_id matches...
  if (session::session_id == session_id) {
    control_netio::send_session_connect_ack(session_id, peer);
  }
}

void 
notify_session_connect_ack(string ack_session_id, inet_address peer) {
  if (session_id == ack_session_id) {
    LOG_EVENT("session") << "connection request ACK. connection established.";

    control_netio::set_session_info(session_id, peer);
    connected = true;

    notify_syscall_session_connected(session_owner);
    notify_core_session_connected();
  }
}

HORDE_NAMESPACE_END;
