#include <string.h>

#include "horde_log.h"
#include "ippc_server.h"

HORDE_NAMESPACE_BEGIN;

//
// ippc_server
//    accepts sockets for ippc

ippc_server::ippc_server(io_event_loop& loop, string path, 
			 ippc_receive_handler recv_handler,
			 ippc_close_handler close_handler,
			 ippc_flush_handler flush_handler) 
  : io_handler(loop), recv_handler(recv_handler), close_handler(close_handler), 
    flush_handler(flush_handler), path(path), handlers()
{      
  sock = ippc_socket(unix_address(path), nmstl::socket::acceptor);

  set_ioh(sock);
  sock.set_blocking(false);
  want_read(true);
  
  LOG_EVENT("ippc") << "server started on path \'" << path << "\'";
}

ippc_handler* 
ippc_server::create_ippc_handler(int fd, 
				 ippc_receive_handler recv_handler, 
				 ippc_close_handler close_handler,
				 ippc_flush_handler flush_handler) {

  if (recv_handler == NULL) recv_handler = this->recv_handler;
  if (close_handler == NULL) close_handler = this->close_handler;
  if (flush_handler == NULL) flush_handler = this->flush_handler;

  ippc_handler* t = new ippc_handler(*get_loop(), ippc_socket(iohandle(fd)), 
				     recv_handler, close_handler, flush_handler);

  handlers.push_back(t);

  return t;
}

void 
ippc_server::ravail() {
  address a;
  ippc_socket client_sock = sock.accept(a);
  if (client_sock) {
    LOG_EVENT("ippc") << "accepted socket on FD#" << client_sock.get_fd() << " (" << client_sock.getpeername() << ")";
    ippc_handler* t = new ippc_handler(*get_loop(), client_sock, recv_handler, close_handler, flush_handler);

    handlers.push_back(t);
  } else {
    LOG_EVENT("ippc") << "CRITICAL: socket accept failed: " << strerror(errno) << ", errno=" << errno;
    stop_server();
  }
}

void 
ippc_server::stop_server(bool call_close_handler) {
  LOG_EVENT("ippc") << "shutting down IPPC server.";
  want_read(false);
  unlink(path.c_str());
  close();

  if (call_close_handler)
    close_handler(NULL); // this guy may call exit...
}

//
// ippc_handler
//    handles messages on ippc sockets
//

// --- READING ---

void
ippc_handler::ravail()  {  
  int type;
  size_t length;

  char *msg = ippc_message_receive(ioc, &type, &length);
  
  if (msg == IPPC_RECV_INCOMPLETE) { // partial read of multi-part
    // do nothing...
  } else if (msg == NULL) {    // error
    if (errno != EAGAIN) {     // ignore async I/O errors...
      LOG_EVENT("ippc") << "socket read failed: " << strerror(errno) << ", errno=" << errno;
      LOG_EVENT("ippc") << "closing socket.";

      close();
    }
  } else { // have a valid message
    recv_handler(this, type, msg, length);
  }
}


// --- WRITING ---
void
ippc_handler::wavail() {
  if (send_queue.size() > 0) {
    send_queue_element* e = send_queue.front();

    ssize_t n = ippc_message_send(ioc, e->type, e->msg.data(), e->msg.length());
    
    if (n == e->msg.length()) {
      send_queue.pop_front(); // done with this message
      delete e;
      LOG_EVENT_DEBUG("ippc") << "ippc_message_send: " << n << " bytes sent, " << send_queue.size() << " messages in send_queue.";

      if (send_queue.empty()) flush_handler(this);
    } else { // error... (n < 0)    
      if (errno == E2BIG) {
	LOG_EVENT_DEBUG("ippc") << "socket write failed: message too big [" << e->msg.length() << "]";
	send_queue.pop_front(); // dump message  
	delete e;
      } else if (errno == EAGAIN) {
	LOG_EVENT_DEBUG("ippc") << "socket write failed: errno=EAGAIN";
	return; // @TODO: what if we get loads of consecutive EAGAIN errors...?
	        //        this'll become a busy-wait loop... which will suck.
      } else { // any other errors are treated as critical errors.
	LOG_EVENT("ippc") << "socket write failed: " << strerror(errno) << ", errno=" << errno;
	LOG_EVENT("ippc") << "closing socket.";

	close();
	return;
      }
    }
  }

  if (send_queue.size() == 0) {
    want_write(false);
  }
}

void ippc_handler::close(bool call_close_handler) {
  want_read(false);
  want_write(false);
  io_handler::close();

  if (call_close_handler)
    close_handler(this); // someone may want to react to this close...
};

// nop event handlers
void nop_recv_handler(ippc_handler* src, int type, char* msg, size_t msglen) {
  // do nothing
}

void nop_close_handler(ippc_handler*) {
  // do nothing
}

void nop_flush_handler(ippc_handler*) {
  // do nothing
}

HORDE_NAMESPACE_END;
