/** @file syscall.h */ 
#ifndef __HORDE_IPPC_SYSCALL_H__
#define __HORDE_IPPC_SYSCALL_H__

#include "shared/objectives.h"

#define SESSION_ID_MAX_LENGTH 63
#define STREAM_ID_MAX_LENGTH 63
#define EXCEPT_MSG_MAX_LENGTH 127
#define NET_ADDRESS_MAX_LENGTH 255
#define CHANNEL_ID_MAX_LENGTH 31
#define STAT_MAX_CHANNELS 16

#ifdef __cplusplus
#include <string.h>
#endif

void debug_touch_data(const char* buf, size_t length);

enum ippc_message_type {
  IPPC_SESSION_COMMAND=0x01,
  IPPC_SESSION_EVENT=0x02,  

  IPPC_STREAM_COMMAND=0x03,  
  IPPC_STREAM_EVENT=0x04,

  IPPC_DATA_SEND=0x05,
  IPPC_DATA_RECV=0x06,
  IPPC_DATA_FEEDBACK=0x07,

  IPPC_STATISTICS=0x10,

  IPPC_PING=0xf7,
  IPPC_PONG=0x7f,

  IPPC_NOP=0x0,
  IPPC_EXCEPTION=0xff
};

enum ippc_exception_code {
  IPPC_EX_BAD_IPPC=1,
  IPPC_EX_ILLEGAL_OPERATION=2,
  IPPC_EX_INSUFFICIENT_PERMISSION=3,
  IPPC_EX_BAD_SESSION_ID=4,
  IPPC_EX_BAD_STREAM_ID=5,
  IPPC_EX_INTERNAL=255
};

struct ippcmsg_nop {
  // empty
};

struct ippcmsg_pingpong {
  int seq;
};

struct ippcmsg_exception {
  int code;
  char msg[EXCEPT_MSG_MAX_LENGTH+1];
};

struct ippcmsg_session_command {
  enum command_t {ACCEPT, CONNECT, JOIN, CLOSE, ENABLE_STATISTICS, DISABLE_STATISTICS} command;

  char session_id[SESSION_ID_MAX_LENGTH+1];
  char peer_address[NET_ADDRESS_MAX_LENGTH+1]; // only for CONNECT
};

struct ippcmsg_session_event {
  enum event_t {CONNECTED, CLOSED} event;

  char session_id[SESSION_ID_MAX_LENGTH+1];
};

struct ippcmsg_stream_command {
  enum command_t {ACCEPT, CONNECT, CLOSE, SET_EVENT_MASK, CLEAR} command;

  char stream_id[STREAM_ID_MAX_LENGTH+1];
  int stream_uid;

  // mask codes for SET_EVENT_MASK
  int event_mask_ack;      // true/false
  int event_mask_loss;     // true/false
  int event_mask_recv;     // true/false
  int event_mask_choke;    // true/false
  int event_mask_flushed;  // true/false
};

struct ippcmsg_stream_event {
  enum event_t {CONNECTED, CLOSED, FLUSHED} event;

  char stream_id[STREAM_ID_MAX_LENGTH+1]; // for connected and closed
  int stream_uid;
};

struct ippcmsg_data_feedback {
  enum feedback_t {ACK, LOSS} type;

  int stream_uid;

  int adu_id;
  int frag_start;
  int frag_length; 
  int rtt_ms; // only for ACK
};

struct ippcmsg_data_send {
  int    hstream;
  int    adu_id;
  size_t data_length;

  bool is_fragment;
  int adu_offset;
  size_t adu_length;

  horde_objective_tags tags;

  char data[1];
};

struct ippcmsg_data_recv {
  int hstream;
  int adu_id;
  int total_length;
  int frag_start;
  int frag_length;

  char data[1];
};

struct ippc_channel_statistics {
  char channel_id[CHANNEL_ID_MAX_LENGTH+1];

  int send_rate;
  int recv_rate;
  int allowed_send_rate;

  int loss_rate;
  int ave_rtt;
};

struct ippcmsg_statistics {
  int total_send_rate;
  int total_recv_rate;
  int total_allowed_send_rate;
  int ave_rtt; // smoothed RTT over all channels (weighted by allowed rate)
  int loss_rate; // loss rate over all channels.

  int nchannels;
  ippc_channel_statistics channel_stats[STAT_MAX_CHANNELS];
};

/**
 * Syscall notes:
 *
 * IPPC_SESSION_COMMAND::{JOIN, ACCEPT, CONNECT}
 *   results in a IPPC_SESSION_EVENT::CONNECTED or an IPPC_EXCEPTION
 * 
 */

#endif
