// -*- mode: C++; -*-

/*
 * NMSTL, the Networking, Messaging, Servers, and Threading Library for C++
 * Copyright (c) 2002 Massachusetts Institute of Technology
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <nmstl/thread>

#include <signal.h>

#undef NMSTL_DEBUG_FILE_H
#define NMSTL_DEBUG_FILE_H "nmstl/thread"

NMSTL_NAMESPACE_BEGIN;

thread::thread_map& thread::threads() {
    static thread_map thr;
    return thr;
}

mutex& thread::threads_lock() {
    static mutex mut;
    return mut;
}

void *thread::launch(void *pthis) {
    thread *th = static_cast<thread *>(pthis);

    pthread_t self = pthread_self();
    if (self)
        th->pth = self;

    locking (threads_lock()) {
        threads()[self] = th;
    }

    th->run();
    th->state = DONE;
    if (th->should_delete) {
        th->delete_self_ok = true;
        delete th;
    }
    return 0;
}

thread::thread(string name) :
    state(NONE), detached(false), destroyed(false),
    should_delete(false), delete_self_ok(false), name(name)
{
    if (name == "main") return;

    locking (threads_lock()) {
        // Make sure the static variables are initialized by the first
        // thread() constructor (at which time no other threads can be
        // running)

        if (threads().empty())
            threads()[pthread_self()] = &first_thread();
    }

    INFO << "Created new thread";
}

thread::~thread() {
    INFO << "Destroying thread " << *this;

    ASSERT(!destroyed);
    
    if (!detached && !delete_self_ok && pthread_self() == pth) {
        COREDUMP << "Thread " << thread::self() << " tried to destroy itself";
    } else if (!detached && (state == RUNNING || state == DONE)) {
        join();
    } else {
        DEBUG << "(no need to join " << *this << ")";
        // Already joined
    }
    destroyed = true;

    locking (threads_lock())
        threads().erase(pth);
}

void thread::start() {
    ASSERT(!destroyed);
    ASSERT(state == NONE);
    state = RUNNING;
    int ret = pthread_create(&pth, 0, &thread::launch, this);
    ASSERT(ret == 0);
}

void thread::join() {
    ASSERT(!destroyed);
    ASSERT(state == RUNNING || state == DONE);

    INFO << "Joining thread " << *this << "...";

    void *thread_ret;
    int ret = pthread_join(pth, &thread_ret);
    ASSERT(ret == 0);

    INFO << "Joined thread " << *this;

    state = JOINED;
}

void thread::detach() {
    ASSERT(!destroyed);
    ASSERT(!detached);

    pthread_detach(pth);
    detached = true;
}

/// Sleeps for a certain duration.
void thread::sleep(ntime duration) {
    struct timespec req = duration.to_timespec(), rem;
    while (1) {
        int ret = nanosleep(&req, &rem);

        if (!(ret == -1 && errno == EINTR))
            return;

        req = rem;
    }
}

/// Returns a string describing the currently running thread.
thread& thread::self() {
    locking (threads_lock()) {
        pthread_t pthself = pthread_self();
        thread_map::const_iterator i = threads().find(pthself);

        if (i == threads().end()) {
            return first_thread();
        } else
            return *(i->second);
    }
}

class dummy_thread : public thread {
public:
    dummy_thread(string name = string()) : thread(name) {}

protected:
    void run() {}
};

thread& thread::first_thread() {
    static dummy_thread first("main");
    static bool init = false;
    if (!init) {
        init = true;
        first.pth = pthread_self();
        first.delete_self_ok = true;
        first.detached = true;
    }
    return first;
}

#ifdef __FreeBSD__
// Fix for broken pthread library
// http://docs.freebsd.org/cgi/getmsg.cgi?fetch=59008+0+archive/2002/freebsd-stable/20020811.freebsd-stable

extern "C" {
    int _thread_sys_close(int fd);
    int __sys_close(int fd) {
        return _thread_sys_close(fd);
    }
};
#endif

NMSTL_NAMESPACE_END;
