#include "horde_log.h"
#include "packet_io.h"

#include <nmstl/debug>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

HORDE_NAMESPACE_BEGIN

logger horde_net_logger;
logger horde_event_logger;

void panic(const char* msg) {
  CRIT << "panic: " << msg;
  LOG_EVENT("PANIC") << msg << " <PID=" << getpid() << ">";
  logging_close();
  ::exit(EXIT_FAILURE);
}

void exit(const char* msg) {
  LOG_EVENT("EXIT") << msg << " <PID=" << getpid() << ">";
  logging_close();
  ::exit(EXIT_SUCCESS);
}

string get_timestamp() { // from ntime::http_date
  char buf[128];
  ntime now = ntime::now();
  time_t sec = now.to_secs();
  tm t;
  if (!gmtime_r(&sec, &t)) return "";
  
  size_t ret = strftime(buf, sizeof buf, "%d-%b-%Y %T", &t);
  if (ret) {
    int msec = (now.to_msecs() % 1000);
    char bufmsec[8];
    snprintf(bufmsec, sizeof bufmsec, ".%03d", msec);
    
    string ret(buf);
    ret += bufmsec;
    ret += " GMT";

    return ret;
  }
  else {
    // should never happen!
    return "";
  }
}

string get_net_timestamp() {
  time_t now = ntime::now().to_msecs();
  return to_string(now);
}

string get_logfilename_postfix() {
  char buf[128];
  ntime now = ntime::now();
  time_t sec = now.to_secs();
  tm t;
  if (!gmtime_r(&sec, &t)) return "";
  
  size_t ret = strftime(buf, sizeof buf, "%d.%b.%Y.%H.%M.%S", &t);
  if (ret)
    return buf;
  else
    return "";
}

bool 
logger::operator ^ (const log_msg& msg) {
  msg.write(file);
  if (LOG_FLUSH) file.flush(); /// @todo will this slow it down too much?

  return true;
}

bool 
logger::operator ^ (const net_log_msg& msg) {
  msg.write(file);
  if (LOG_FLUSH) file.flush();

  return true;
}

void
log_msg::write(ostream& ostream) const {
  ostream << get_timestamp() << "\t";

  if (event_id) {
    // event-id max-len of 10 chars. (align it correctly)
    char out[]="            \0";
    for (int i = 0; i < 10; i++) {
      if (event_id[i] == '\0') break;
      out[i] = event_id[i];
    }
    ostream << out;
  }

  ostream << str;

  if (code_info) {
    ostream << " [" << file << ":" << line << "]";
  }

  ostream << endl;
}

void
net_log_msg::write(ostream& ostream) const {
  ostream << get_net_timestamp();

  if (event_id) {
    // event-id max-len of 8 chars. (align it correctly)
    char out[]="\t        \t\0";
    for (int i = 0; i < 8; i++) {
      if (event_id[i] == '\0') break;
      out[i+1] = event_id[i];
    }
    ostream << out;
  }

  ostream << str;

  if (code_info) {
    ostream << "\t[" << file << ":" << line << "]";
  }

  ostream << endl;
}

net_log_msg& net_log_msg::operator<< (horde_net_packet& packet) {
  // format: <type> <length> <sequence> <timestamp> [rtt]
  packet_header* hdr = packet.get_header();
  *this << (int)packet.get_type() << "\t" << (int)packet.get_length() << "\t" 
	<< (int)hdr->get_sequence() << "\t" << hdr->timestamp;

  return *this;
}

bool
logging_init(const char *prefix, const char *base_path) {  
  string postfix = get_logfilename_postfix();

  string logfilename_event; 
  logfilename_event << base_path << "/logs/event/" << prefix << "." << postfix << ".log";

  string logfilename_net;
  logfilename_net << base_path << "/logs/net/" << prefix << "." << postfix << ".log";

  horde_event_logger.open (logfilename_event.c_str());
  horde_net_logger.open (logfilename_net.c_str());

  if (horde_event_logger.is_open()) {
    LOG_EVENT("log") << "init";
  }

  if (horde_net_logger.is_open()) {
    LOG_NET("log") << "init";
  }

  return (horde_event_logger.is_open() && horde_net_logger.is_open());
}

void
logging_close() {
  if (horde_event_logger.is_open()) {
    LOG_EVENT("log") << "close";
    horde_event_logger.close();
  }

  if (horde_net_logger.is_open()) {
    LOG_NET("log") << "close";
    horde_net_logger.close();
  }
}

void
logging_flush() {
  if (horde_event_logger.is_open()) {
    horde_event_logger.flush();
  }

  if (horde_net_logger.is_open()) {
    horde_net_logger.flush();
  }
}

HORDE_NAMESPACE_END
