// -*- mode: C++; -*-

/** @file packet_io.g
 * 
 * horribly designed and ill-documented packet I/O utility functions
 * used by the horde backend...
 *
 * @todo rewrite this
 */

#ifndef __HORDE_PACKET_IO_H__
#define __HORDE_PACKET_IO_H__

#include <stdint.h>
#include <nmstl/io>
#include <string>

using namespace nmstl;
using namespace std;

#include "horde_namespace.h"
//#include "horde_log.h"

string to_hexstr(const char* data, size_t datalen);

HORDE_NAMESPACE_BEGIN;

enum horde_packet_type {
  PACKET_TYPE_DATA=0x00,
  PACKET_TYPE_CODED_DATA=0x0f,

  PACKET_TYPE_SESSION_CONNECT_REQUEST=0x1,
  PACKET_TYPE_SESSION_CONNECT_ACK=0x2,
  PACKET_TYPE_SESSION_CONNECT_ACK_ACK=0x3,
  PACKET_TYPE_SESSION_CONNECT_NACK=0x4,
  PACKET_TYPE_SESSION_CLOSE_REQUEST=0x5,
 
  PACKET_TYPE_STREAM_CONNECT_REQUEST=0x6,
  PACKET_TYPE_STREAM_CONNECT_ACK=0x7,
  PACKET_TYPE_STREAM_CONNECT_NACK=0x8,
  PACKET_TYPE_STREAM_CLOSE_REQUEST=0x9,

  PACKET_TYPE_COMPOSITE_ACK=0xa,

  PACKET_TYPE_INTERFACE_EVENT=0xb,

  PACKET_TYPE_COMPOSITE_CONTROL=0xc
};

inline string packet_type_to_string(horde_packet_type t) {
  switch (t) {
  case PACKET_TYPE_DATA:
    return "DATA";
  case PACKET_TYPE_CODED_DATA:
    return "CODED_DATA";
  case PACKET_TYPE_SESSION_CONNECT_REQUEST:
    return "SESSION_CONNECT_REQUEST";
  case PACKET_TYPE_SESSION_CONNECT_ACK:
    return "SESSION_CONNECT_ACK";
  case PACKET_TYPE_SESSION_CONNECT_ACK_ACK:
    return "SESSION_CONNECT_ACK_ACK";
  case PACKET_TYPE_SESSION_CONNECT_NACK:
    return "SESSION_CONNECT_NACK";
  case PACKET_TYPE_SESSION_CLOSE_REQUEST:
    return "SESSION_CLOSE_REQUEST";
  case PACKET_TYPE_STREAM_CONNECT_REQUEST:
    return "STREAM_CONNECT_REQUEST";
  case PACKET_TYPE_STREAM_CONNECT_ACK:
    return "STREAM_CONNECT_ACK";
  case PACKET_TYPE_STREAM_CONNECT_NACK:
    return "STREAM_CONNECT_NACK";
  case PACKET_TYPE_STREAM_CLOSE_REQUEST:
    return "STREAM_CLOSE_REQUEST";
  case PACKET_TYPE_COMPOSITE_ACK:
    return "COMPOSITE_ACK";
  case PACKET_TYPE_INTERFACE_EVENT:
    return "INTERFACE_EVENT";
  case PACKET_TYPE_COMPOSITE_CONTROL:
    return "COMPOSITE_CONTROL";
  default:
    return string("UNKNOWN:")+((int)t);
  }
}
  
struct packet_header {
  // filled in by channel manager (64-bit channel manager header)
  uint32_t channel_id_seq; // bits 24-31: channel_id; bits 0-23: sequence
  uint32_t timestamp;

  // general fields (8-bit general header)
  uint8_t  type;        // bits 0-3: type identifier; 4-7: flags

  // util methods
  uint8_t get_channel_id() const { 
    return (channel_id_seq & 0xff000000) >> 24;
  }

  uint32_t get_sequence() const {
    return (channel_id_seq & 0x00ffffff);
  }

  void set_channel_id (uint8_t id) {
    channel_id_seq = (channel_id_seq & 0x00ffffff) | (id << 24);
  }

  void set_sequence(uint32_t seq) {
    channel_id_seq = (seq & 0x00ffffff) | (channel_id_seq & 0xff000000);
  }
};

struct packet_ack {
  uint32_t channel_id_seq; // bits 24-31: channel_id; bits 0-23: sequence
  uint32_t timestamp_echo;

  int get_channel_id() const {
    return (channel_id_seq & 0xff000000) >> 24;
  }

  uint32_t get_sequence()  const {
    return (channel_id_seq & 0x00ffffff);
  }
};

class horde_net_packet {
private:
  dynbuf databuf;

  // header fields:
  //
  //     0: channel_id            (8-bit)   +1
  //   1-3: channel_seq           (24-bit)  +3
  //   4-7: timestamp             (32-bit)  +4
  //
  //     8: packet type and flags (8-bit)   +1
  //
  //                                        = 9
  //
  // note: not the most efficient storage, but, it works...    

public:
  static const size_t header_sz = sizeof(packet_header);

  horde_net_packet(const horde_net_packet& other) : databuf(other.databuf) { 
    //    LOG_EVENT_DEBUG("debug") << "copied packet buffer: " << (void*)databuf.data()
    //			     << ", refcount:" << databuf.get_refcount();
  };

  horde_net_packet(const dynbuf& data) : databuf(data) { 
    //    LOG_EVENT_DEBUG("debug") << "copied packet buffer: " << (void*)databuf.data()
    //			     << ", refcount:" << databuf.get_refcount();
  }

  horde_net_packet() : databuf() { }

  ~horde_net_packet() { 
    //    LOG_EVENT_DEBUG("debug") << "releasing packet buffer: " << (void*)databuf.data()
    //			     << ", refcount:" << databuf.get_refcount();
  }

  /// for valid packets, requires: (buf_sz >= header_sz)
  horde_net_packet(size_t buf_sz, horde_packet_type t=PACKET_TYPE_DATA) : databuf(buf_sz) {       
    packet_header* header = get_header();
    memset((void*)header, 0, header_sz);
    
    set_type(t);
    set_length(buf_sz);

    //    LOG_EVENT_DEBUG("debug") << "allocated new packet buffer: " << (void*)databuf.data()
    //			     << ", refcount:" << databuf.get_refcount();
  }

  horde_net_packet& operator= (const horde_net_packet& other) {
    databuf = other.databuf;

    //    LOG_EVENT_DEBUG("debug") << "copied packet buffer: " << (void*)databuf.data()
    //			     << ", refcount:" << databuf.get_refcount();
  }
  
  dynbuf& get_buf() {
    return databuf;
  }
    
  packet_header* get_header() {
    //return (uint8_t)(*(databuf.data()));
    return ((packet_header*)(databuf.data()));
  }

  packet_ack* get_ack() {
    return ((packet_ack*)(databuf.data()));
  }

  horde_packet_type get_type()  {
    int x = ((packet_header*)(databuf.data()))->type;
    return (horde_packet_type)(x & 0x0f);
  }

  void set_type (horde_packet_type t) {
    ((packet_header*)(databuf.data()))->type = 
      (((packet_header*)(databuf.data()))->type & 0xf0) | (int8_t(t) & 0x0f);      
  }

  int get_flags()  {
    return ((((packet_header*)(databuf.data()))->type & 0xf0) >> 4);
  }

  void set_flags(int new_flags) {
    new_flags = (new_flags & 0x0f) << 4;

    ((packet_header*)(databuf.data()))->type = 
      (((packet_header*)(databuf.data()))->type & 0x0f) | new_flags;      
  }

  //
  // packet payload stuff
  // 

  /// size of the packet minus the (64-bit) header
  size_t get_data_size()  {
    return databuf.length() - header_sz;
  }

  size_t get_length()  {
    return databuf.length();
  }

  /// new_sz must include header_sz
  /// requires: new_sz >= header_sz
  size_t set_length(size_t new_sz) {
    databuf.set_length(new_sz);
  }

  /// returns a pointer to the underlying dynbuf's data
  char* get_data() {
    return (databuf.data() + header_sz);
  }

  string to_debug_string()  {
    string retval = "packet { type=";
    retval += packet_type_to_string(get_type());
    retval += ", length=";
    retval += (int)get_length();
    retval += ", channel_seq=";
    retval += (int)get_header()->get_sequence();

    switch (get_type()) {
    case PACKET_TYPE_SESSION_CONNECT_REQUEST:
    case PACKET_TYPE_SESSION_CONNECT_ACK:
    case PACKET_TYPE_SESSION_CONNECT_ACK_ACK:
    case PACKET_TYPE_SESSION_CONNECT_NACK:
    case PACKET_TYPE_SESSION_CLOSE_REQUEST:
      {
	// assume the payload is a string
	char *x = get_data();
	retval += ", session=";
	retval += x;	
      }
      break;
    case PACKET_TYPE_STREAM_CONNECT_REQUEST:
    case PACKET_TYPE_STREAM_CONNECT_ACK:
    case PACKET_TYPE_STREAM_CONNECT_NACK:
    case PACKET_TYPE_STREAM_CLOSE_REQUEST:
      {
	// assume the payload is a string
	char *x = get_data();
	retval += ", stream=";
	retval += x;	
      }
      break;
    case PACKET_TYPE_DATA:
      {
	retval += ", payload: ";
	size_t n = get_data_size();
	if (n > 16) n = 16;
	retval += to_hexstr(get_data(), n);
      }
      break;
    case PACKET_TYPE_COMPOSITE_ACK:      
      {
	packet_ack* ack = get_ack();
	retval += ", ack_sequence: ";
	retval += ack->get_sequence();
      }
      break;      
    }

    retval += '}';
    
    return retval;
  }

  string to_log_string() {
    return to_debug_string(); ///@todo implement logging functionality
  }
};

HORDE_NAMESPACE_END;

#endif
