/**
 * Program to run horde bandwidth tests. 
 *
 * Client connects to server and uploads ADUs. 
 * 
 * Client produces a statistics log on standard error. The log contains info
 * on overall statistics as the first group of columns, followed by
 * info for each channel active at that point. Each group of columns
 * consists of: 
 * 'group_name recv_rate send_rate allowed_send_rate loss_rate smoothed_rtt'
 *
 * group_name is TOTAL for the first column, and the channel identifier for
 * every other group.
 *
 * Also an example of the horde API.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <iostream>
#include <iostream>
#include <limits>
#include <fstream>
#include <set>
#include <cmath>
#include <ctime>
#include <cassert>
#include <string>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>

#include <string>
using namespace std;

#include "horde.h"

// 
// globals
//

char *sessionid = "ping";
char *streamid = "ping";
int hstream = -1;

//
// utility functions
//

int horde_event_loop(int timeout=-1) {
  int k = 0;
  while ((k = horde_event_dispatch(timeout)) > 0) ;
  return k;
}

//
// send/recv
//

int seq = 0;

void send_ping() {
  char *buf = (char*)calloc(32,1);
  snprintf(buf, 32, "PING:%d", seq);
  seq++;
  
  horde_objective_tags tags;
  horde_stream_send(hstream, seq, buf, 32, tags);
  
  cout << "send -> " << buf << endl;
}

void stream_event_cb(horde_stream_event event_code, int stream_uid, int adu_id, int offset, int len, int rtt, void* unusedptr) {
  if (event_code == ADU_FRAG_LOSS) {
    fprintf(stderr, "\tLOSS: adu=%d\n", adu_id);

    // retransmit ping
    send_ping();
  } else if (event_code == ADU_FRAG_ACK) {
    fprintf(stderr, "\tACK: adu=%d, rtt=%d ms\n", adu_id, rtt);
  }
}

void stream_recv_cb (int stream, int adu_id, int total_len, int frag_offset, int frag_len, const char* frag_data, void* unused) {
  fprintf(stderr, "\tRECV: adu=%d, len=%d\n", adu_id, frag_len);

  cout << "recv <- " << frag_data << endl;
  
  send_ping();
}

//
// server
//
void run_server() {
  cout << "Waiting for client...";
  if (horde_session_accept(sessionid)) {
    cout << "session up...";
    hstream = horde_stream_accept(streamid);
    if (hstream > 0) {
      cout << "stream connected [" << hstream  << "]\n";

      horde_stream_set_event_callback(hstream, stream_event_cb, NULL);
      horde_stream_set_recv_callback(hstream, stream_recv_cb, NULL);

      // enter event loop
      horde_event_loop();

      // done
      horde_stream_close(streamid, hstream);
    }
    
    horde_session_close(sessionid);  
  }
}

//
// client
//

void run_client(char* server) {
  cout << "Contacting server...";
    if (horde_session_connect(server, sessionid)) {
    cout << "session up...";
    hstream = horde_stream_connect(streamid);
    if (hstream > 0) {
      cout << "stream connected [" << hstream  << "]\n";

      horde_stream_set_event_callback(hstream, stream_event_cb, NULL);
      horde_stream_set_recv_callback(hstream, stream_recv_cb, NULL);

      send_ping();

      // enter event loop
      horde_event_loop();

      // done
      horde_stream_close(streamid, hstream);
    }
    
    horde_session_close(sessionid);  
  }
}

//
// main
//

int main(int argc, char **argv)
{ 
  if ((argc != 1) && (argc != 2)) {
    cerr << "correct usage: " << argv[0] << " [server address]" << endl;
  } else if (argc == 1) {
    run_server();
  } else {
    run_client(argv[1]);
  }
}
