#include "raw_io.h"
#include "raw_iohandler.h"

using namespace std;
using namespace nmstl;
 
//
// raw socket I/O classes; one for TCP, one for UDP, one for ICMP, 
// and one for sending...
//

// handle UDP
class rawio_udp_iohandler : public filtered_raw_iohandler {
protected:
  void handle_packet(dynbuf& ip_packet) {
    DEBUG(3) << "UDP RECV: " << packet_header_to_string(ip_packet);
    handle_raw_recv(ip_packet);
  }

public:
  rawio_udp_iohandler(io_event_loop& ioloop, inet_address& bind_address) 
    : filtered_raw_iohandler(ioloop, UDP, bind_address, true, false, false)
  { }
};

// handle TCP
class rawio_tcp_iohandler : public filtered_raw_iohandler {
protected:
  void handle_packet(dynbuf& ip_packet) {
    DEBUG(3) << "TCP RECV: " << packet_header_to_string(ip_packet);
    handle_raw_recv(ip_packet);
  }

public:
  rawio_tcp_iohandler(io_event_loop& ioloop, inet_address& bind_address) 
    : filtered_raw_iohandler(ioloop, TCP, bind_address, true, false, false)
  { }
};

// handle ICMP
class rawio_icmp_iohandler : public filtered_raw_iohandler {
protected:
  void handle_packet(dynbuf& ip_packet) {
    DEBUG(3) << "ICMP RECV: " << packet_header_to_string(ip_packet);
    handle_raw_recv(ip_packet);
  }

public:
  rawio_icmp_iohandler(io_event_loop& ioloop, inet_address& bind_address) 
    : filtered_raw_iohandler(ioloop, ICMP, bind_address, true, false, false)
  { } 
};

// for sending...
class rawio_send_iohandler : public filtered_raw_iohandler {
protected:
  void handle_packet(dynbuf& ip_packet) {
    WARN << "unexpected handle_packet in rawio_send : " << packet_header_to_string(ip_packet);
  }

public:
  rawio_send_iohandler(io_event_loop& ioloop, inet_address& bind_address) 
    : filtered_raw_iohandler(ioloop, RAW, bind_address, false, false, false)
  { } 
};

//
// raw socket I/O handlers
//

rawio_udp_iohandler* udp_handler = NULL;
rawio_tcp_iohandler* tcp_handler = NULL;
rawio_icmp_iohandler* icmp_handler = NULL;
rawio_send_iohandler* send_handler = NULL;

rawio_handler_group::rawio_handler_group(inet_address& bind_address) {
  udp_handler = new rawio_udp_iohandler(io_loop, bind_address);
  tcp_handler = new rawio_tcp_iohandler(io_loop, bind_address);
  icmp_handler = new rawio_icmp_iohandler(io_loop, bind_address);
  send_handler = new rawio_send_iohandler(io_loop, bind_address);

  DEBUG << "created raw I/O handlers: TCP:" << tcp_handler->is_valid()
        << ", UDP:" << udp_handler->is_valid()         
	<< ", ICMP:" << icmp_handler->is_valid()
        << ", SEND:" << send_handler->is_valid();
}

void 
rawio_handler_group::close() {
  // UDP
  if (udp_handler) {
    udp_handler->close();
    delete udp_handler;
    udp_handler = NULL;
  }

  // TCP
  if (tcp_handler) {
    tcp_handler->close();
    delete tcp_handler;
    tcp_handler = NULL;
  }

  // ICMP
  if (icmp_handler) {
    icmp_handler->close();
    delete icmp_handler;
    icmp_handler = NULL;
  }

  // SEND
  if (send_handler) {
    send_handler->close();
    delete send_handler;
    send_handler = NULL;
  }

  DEBUG << "destroyed raw I/O handlers.";
}

rawio_handler_group::~rawio_handler_group() {
  if (udp_handler || tcp_handler || icmp_handler || send_handler) {
    close();
  }
}


//
// other exported methods...
//

void 
rawio_handler_group::send_raw(dynbuf& packet) {
  if (send_handler) {    
    //packet.set_offset(0);
    //short *pdata = (short *)packet.data();
    //pdata[0] = 0x0000;
    //pdata[1] = 0x0008;

    int n = send_handler->send(packet);

    if (n < 0) {
      WARN << "send_raw: error n=" << n << ", errno=" << errno << ", len=" << packet.length();
    } else {
      DEBUG(3) << "send_raw: sendto n=" << n;
    }
  }
  else 
    WARN << "send_raw called without a send I/O handler.";
}

void 
rawio_handler_group::allow_address(const inet_address& addr) {
  if (udp_handler)  udp_handler->add_allowed_dest(addr);
  if (tcp_handler)  tcp_handler->add_allowed_dest(addr);
  if (icmp_handler) icmp_handler->add_allowed_dest(addr);
}

void 
rawio_handler_group::ignore_address(const inet_address& addr) {
  if (udp_handler)  udp_handler->remove_allowed_dest(addr);
  if (tcp_handler)  tcp_handler->remove_allowed_dest(addr);
  if (icmp_handler) icmp_handler->remove_allowed_dest(addr);
}
