// -*- mode: C++; -*-

/** @file properties.h 
 *
 * Utility functions to manage property lists, lists of string
 * name/value pairs, in strings and files. Both names and values are
 * case sensitive
 * 
 * <b>String I/O</b>
 * Strings of the form \c "x=1,y=2,z=3" can be converted into property
 * sets. See \a properties::parse_parameter_string().
 * 
 * <b>File I/O</b>
 * Properties can be loaded from, and saved to, files. The file format is
 * a little different from the <i>parameter string</i> format. For example:
 *
 * \code
 * # Debug logging level
 * log_level: 5
 *
 * # Managed interfaces
 * managed_interfaces: ppp0, tun0, tun1
 * \endcode
 *
 * Any line beginning with a \c \# is ignored as a comment.  Other
 * lines must begin with a property name, followed by a value.  A
 * colon (\c :) is used to seperate property names and values. Leading
 * and trailing whitespace is trimmed from the value. If the same name
 * occurs twice in a file, only the later value is used. See \a
 * properties::read_from_file() and \a properties::write_to_file().
 *
 *
 * <b>Value Lists</b>
 * There are also some utility functions to convert 'arrays', or value lists
 * (e.g., the string <tt>"ppp0, tun0, tun1"</tt>), to and from string
 * vectors (e.g., <tt>{"ppp0", "tun0", "tun1"}</tt>). See \a
 * properties::decode_array_value() and \a
 * properties::encode_array_value().
 *
 * @todo Clean up implementation and document properly.
 * @author Asfandyar Qureshi (asfand)
 */

/*
 * Old description... is this still accurate?
 *
 * <newline> := the newline character
 * <any_char> := any character other than newline
 * <config_file> := <lines>*
 * <line> := <config_file_element> <newline>
 * <config_file_element> := <comment> | <name_value_pair>
 * <comment> := '#' <any_char>*
 * <name_value_pair> := <name> '=' <value>
 *    (so, leading/trailing spaces are included in the value, but not in the name)
 * <name> := ([a-zA-z0-9] | underscore_character)+
 */

//
// Config file I/O utility functions...
//

#ifndef __HORDE_CONFIG_FILE_H__
#define __HORDE_CONFIG_FILE_H__

#include "string_hashmap.h"

string trim_spaces(const string& input);

string to_lowercase(const string& input);
string to_uppercase(const string& input);

/**
 * A set of name/value pairs. Supports parsing-from and writing-to files.
 */
class properties {
 protected:
  //typedef string_string_hashmap_t hash_map_t;
  typedef STRING_HASHMAP(string) hash_map_t;
  hash_map_t property_set;
 public:
  /**
   * Constructs an empty set of properties.
   */
  properties() : property_set() { }

  /**
   * Construct a set by parsing a parameter string.
   *
   * @see parse_parameter_string
   */
  properties(const string& paramstring) : property_set() {
    parse_parameter_string(paramstring);
  }

  /**
   * Cleans up memory footprint.
   */
  virtual ~properties() { /* nothing to do */ }

  /**
   * Parses a simple parameter string of the following form:
   *
   * \code
   * "x=1,y=2,z=3"
   * \endcode
   *
   * Into the property set: 
   *
   * \code
   * { "x" -> "1", "y" ->"2", "z" -> "3" }
   * \endcode
   *
   * The properties in the string are added to this property set
   * object, overwriting any older properties with  the same names.
   *
   * @note This really belongs somewhere else. Rfactor later..  I
   *       don't like this here because the format of paramstring is fairly
   *       different from the rest of the file format...
   */
  virtual int parse_parameter_string(const string& paramstring);

  /**
   * Parses the file and ADDS the properties from that file into the
   * current set of properties. Overwrites older properties if the
   * file contains values for them. If an error occurs during the
   * parsing of the file, then properties before the parse error may
   * or may not have been added to the properties set.
   *
   * @return <0 in case of error, otherwise returns the number of properties
   *    added from the file.
   */
  virtual int read_from_file(const string& filename, bool errors_to_cerr=false); 

  /**
   * Writes the properties to a file. No comments.
   *
   * @return <0 in case of error, otherwise returns the number of properties
   *    written to the file.
   */
  virtual int write_to_file(const string& filename) const;

  
  bool contains(const string& property_name) const;

  /**
   * Get a property's value. 
   *
   * @return "" if property doesn't exist, or the value otherwise.
   */
  string get(const string& property_name) const;

  /**
   * Get a property's value, and insert the given default value
   * into the properties set if the value does not exist.
   *
   * @return returns default_value if property doesn't exist, or the
   *    value otherwise.
   */
  string get(const string& property_name, const string& default_value);
  
  /**
   * Set a property value. 
   *
   * @param value  The impl of this method should create its own copy of this...
   * @param overwrite_existing  Specifies whether this set should overwrite an 
   *                            older value for the property.
   */
  void set(const string& property_name, const string& value, bool overwrite_existing=true);

  /**
   * Lists all the property names.
   */
  vector<string> list_properties() const;

  /**
   * Removes all properties from this set.
   */
  void clear() { property_set.clear(); }

  /**
   * String representation.
   */
  string to_string() const;

  /**
   * String representation with inferred value types. Useful for debugging this class.
   */
  string to_string_x() const;

  /**
   * Returns the number of keys.
   */
  size_t size() const { return property_set.size(); }

  //
  // utility functions
  //

  /**
   * Converts a comma seperated string into an array.
   *
   * @note this really belongs somewhere else. refactor later...
   */
  static vector<string> decode_array_value(const string& value);
  static string encode_array_value(vector<string>);

  enum property_value_type { 
    TYPE_UNKNOWN = 0, 
    TYPE_STRING_ARRAY = 1,
    TYPE_STRING = 2,
    TYPE_INTEGER_ARRAY = 3,
    TYPE_INTEGER = 4,
    TYPE_FLOAT_ARRAY = 5,
    TYPE_FLOAT = 6,
    ARRAY_TYPE_MASK = 0x01
  };

  /**
   * Try to guess the 'type' of a string.
   */
  static property_value_type infer_value_type(const string& value);   

  /**
   * Get a human readable type description.
   */
  static const char* get_vtype_description(property_value_type t);
};

#endif
